"""
This module implements the SOMO method.
"""

import numpy as np

from scipy.linalg import circulant

import minisom

from ..base import pairwise_distances_mahalanobis

from ..base import coalesce, coalesce_dict, fix_density
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['SOMO']

class SOMO(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @article{somo,
                        title = "Self-Organizing Map Oversampling (SOMO) for
                                    imbalanced data set learning",
                        journal = "Expert Systems with Applications",
                        volume = "82",
                        pages = "40 - 52",
                        year = "2017",
                        issn = "0957-4174",
                        doi = "https://doi.org/10.1016/j.eswa.2017.03.073",
                        author = "Georgios Douzas and Fernando Bacao"
                        }

    Notes:
        * It is not specified how to handle those cases when a cluster contains
            1 minority samples, the mean of within-cluster distances is set to
            100 in these cases.
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_uses_clustering,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 *,
                 n_grid=10,
                 sigma=0.2,
                 learning_rate=0.5,
                 n_iter=100,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_grid (int): size of grid
            sigma (float): sigma of SOM
            learning_rate (float) learning rate of SOM
            n_iter (int): number of iterations
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, 'proportion', 0)
        self.check_greater_or_equal(n_grid, 'n_grid', 2)
        self.check_greater(sigma, 'sigma', 0)
        self.check_greater(learning_rate, 'learning_rate', 0)
        self.check_greater_or_equal(n_iter, 'n_iter', 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.params = {'n_grid': n_grid,
                        'sigma': sigma,
                        'learning_rate': learning_rate,
                        'n_iter': n_iter}
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_grid': [5, 9, 13],
                                  'sigma': [0.4],
                                  'learning_rate': [0.3, 0.5],
                                  'n_iter': [100]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def construct_grid(self, X, y):
        """
        Construct the grid.

        Args:
            X (np.array): all feature vectors
            y (np.array): all target labels

        Returns:
            dict, dict: the minority and majority grids
        """
        # training SOM
        som = minisom.MiniSom(self.params['n_grid'],
                              self.params['n_grid'],
                              len(X[0]),
                              sigma=self.params['sigma'],
                              learning_rate=self.params['learning_rate'],
                              random_seed=3)
        som.train_random(X, self.params['n_iter'])

        # constructing the grid
        grid_min = {}
        grid_maj = {}

        for jdx, x_vec in enumerate(X):
            idx = som.winner(x_vec)
            idx = (idx[0], idx[1])
            if idx not in grid_min:
                grid_min[idx] = []
            if idx not in grid_maj:
                grid_maj[idx] = []
            if y[jdx] == self.min_label:
                grid_min[idx].append(jdx)
            else:
                grid_maj[idx].append(jdx)

        # converting the grid to arrays
        for idx in grid_min:
            grid_min[idx] = np.array(grid_min[idx])
        for idx in grid_maj:
            grid_maj[idx] = np.array(grid_maj[idx])

        return grid_min, grid_maj

    def determine_densities(self, grid_min, grid_maj, X, y):
        """
        Determine densities

        Args:
            grid_min (dict): the minority grid
            gird_maj (dict): the majority grid
            X (np.array): all feature vectors
            y (np.array): all target labels

        Returns:
            dict: the densities
        """
        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= \
                    self.metric_tensor_from_nn_params(nn_params, X, y)

        tensor = nn_params.get('metric_tensor', None)

        # filtering
        filtered = {}
        for idx, _ in grid_min.items():
            filtered[idx] = (len(grid_maj[idx]) + 1)/(len(grid_min[idx]) + 1) < 1.0

        # computing densities
        densities = {}
        for idx, filt in filtered.items():
            if filt:
                paird = pairwise_distances_mahalanobis(X[grid_min[idx]],
                                                            tensor=tensor)
                mean_value = np.mean(paird)
                if len(grid_min[idx]) > 1 and mean_value > 0:
                    densities[idx] = len(grid_min[idx])/mean_value**2
                else:
                    densities[idx] = 10

        return densities

    def determine_pair_densities(self, densities):
        """
        Determine pair densities.

        Args:
            densities (dict): the densities

        Returns:
            dict: the pair densities
        """
        # computing neighbour densities, using 4 neighborhood
        neighbors = [[0, 1], [0, -1], [1, 0], [-1, 0]]
        pair_densities = {}
        for idx, _ in densities.items():
            for neigh in neighbors:
                jdx = (idx[0] + neigh[0], idx[1] + neigh[1])
                if jdx in densities:
                    pair_densities[(idx, jdx)] = densities[idx] + densities[jdx]

        return pair_densities

    def create_density_from_dict(self, density):
        """
        Create probability density from a dictionary.

        Args:
            density (dict): densities unnormalized in a dictionary

        Returns:
            np.array: the density values
        """
        density_vals = np.array(list(density.values()))
        density_vals = 1.0 / density_vals

        density_vals = fix_density(density_vals)

        return density_vals

    def generate_intra(self, *,
                        density_keys,
                        density_vals,
                        grid_min,
                        X,
                        n_to_sample):
        """
        Generate samples within the clusters.

        Args:
            density_keys (list): the density keys
            density_vals (np.array): the density values
            grid_min (dict): the minority grid
            X (np.array): all feature vectors
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """
        samples = np.zeros((0, X.shape[1]))
        clusters = self.random_state.choice(np.arange(len(density_keys)),
                                            n_to_sample,
                                            p=density_vals)

        cluster_unique, cluster_count = np.unique(clusters, return_counts=True)

        for idx, cluster_idx in enumerate(cluster_unique):
            cluster = X[grid_min[density_keys[cluster_idx]]]
            #n_dim_orig = self.n_dim
            #self.n_dim = np.min([len(cluster), n_dim_orig])
            samples_tmp = self.sample_simplex(
                                X=cluster,
                                indices=circulant(np.arange(cluster.shape[0])),
                                n_to_sample=cluster_count[idx])
            samples = np.vstack([samples, samples_tmp])
            #self.n_dim = n_dim_orig

        #while len(samples) < n_to_sample:
        #    cluster_idx = density_keys[self.random_state.choice(
        #        np.arange(len(density_keys)), p=density_vals)]
        #    cluster = grid_min[cluster_idx]
        #    sample_a, sample_b = self.random_state.choice(cluster, 2)
        #    samples.append(self.sample_between_points(
        #        X[sample_a], X[sample_b]))

        return samples

    def generate_inter_samples(self, cluster_a, cluster_b, n_to_sample):
        """
        Generate samples between two clusters.

        Args:
            cluster_a (np.array): the first cluster
            cluster_b (np.array): the second cluster
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """
        tile = np.tile(np.arange(len(cluster_b)), (len(cluster_a), 1))

        indices = (np.vstack([np.zeros(len(cluster_a)), tile.T]).T).astype(int)

        n_dim_orig = self.n_dim
        self.n_dim = 2

        samples_tmp = self.sample_simplex(
                            X=cluster_a,
                            indices=indices,
                            n_to_sample=n_to_sample,
                            X_vertices=cluster_b)

        self.n_dim = n_dim_orig

        return samples_tmp

    def generate_inter(self, *,
                        pair_keys,
                        pair_dens_vals,
                        grid_min,
                        X,
                        n_to_sample):
        """
        Generate samples between the clusters.

        Args:
            pair_keys (list): the density keys
            pair_dens_vals (np.array): the density values
            grid_min (dict): the minority grid
            X (np.array): all feature vectors
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """

        samples = np.zeros((0, X.shape[1]))

        if pair_dens_vals.shape[0] == 0:
            return samples

        clusters = self.random_state.choice(np.arange(len(pair_keys)),
                                            n_to_sample,
                                            p=pair_dens_vals)

        cluster_unique, cluster_count = np.unique(clusters, return_counts=True)

        for idx, cluster_idx in enumerate(cluster_unique):

            samples_tmp = self.generate_inter_samples(
                            cluster_a=X[grid_min[pair_keys[cluster_idx][0]]],
                            cluster_b=X[grid_min[pair_keys[cluster_idx][1]]],
                            n_to_sample=cluster_count[idx]
                            )
            samples = np.vstack([samples, samples_tmp])


        #while len(samples) < n_to_sample:
        #    idx = pair_keys[self.random_state.choice(
        #        np.arange(len(pair_keys)), p=pair_dens_vals)]
        #    cluster_a = grid_min[idx[0]]
        #    cluster_b = grid_min[idx[1]]
        #    X_a = X[self.random_state.choice(cluster_a)]
        #    X_b = X[self.random_state.choice(cluster_b)]
        #    samples.append(self.sample_between_points(X_a, X_b))

        return samples

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion,
                                           self.class_stats[self.maj_label],
                                           self.class_stats[self.min_label])

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        grid_min, grid_maj = self.construct_grid(X, y)

        densities = self.determine_densities(grid_min, grid_maj, X, y)

        # all clusters can be filtered
        if len(densities) == 0:
            return self.return_copies(X, y, "all clusters filtered")

        pair_densities = self.determine_pair_densities(densities)

        density_vals = self.create_density_from_dict(densities)
        pair_dens_vals = self.create_density_from_dict(pair_densities)

        density_keys = list(densities.keys())
        pair_keys = list(pair_densities.keys())

        # determining num of samples to generate
        if len(pair_dens_vals) > 0:
            dens_num = int(n_to_sample / 2)
        else:
            dens_num = n_to_sample

        # generating the samples according to the extracted distributions
        samples_intra = self.generate_intra(density_keys=density_keys,
                                            density_vals=density_vals,
                                            grid_min=grid_min,
                                            X=X,
                                            n_to_sample=dens_num)

        samples_inter = self.generate_inter(pair_keys=pair_keys,
                                            pair_dens_vals=pair_dens_vals,
                                            grid_min=grid_min,
                                            X=X,
                                            n_to_sample=n_to_sample-dens_num)

        return (np.vstack([X, np.vstack([samples_intra,
                                         samples_inter])]),
                np.hstack([y, np.repeat(self.min_label,
                                        len(samples_intra) \
                                            + len(samples_inter))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_grid': self.params['n_grid'],
                'sigma': self.params['sigma'],
                'learning_rate': self.params['learning_rate'],
                'n_iter': self.params['n_iter'],
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the DEAGO method.
"""

import warnings
warnings.simplefilter("ignore", DeprecationWarning)

import numpy as np # pylint: disable=wrong-import-position
from sklearn.preprocessing import StandardScaler # pylint: disable=wrong-import-position

import tensorflow # pylint: disable=wrong-import-position

from tensorflow.keras.layers import Input, Dense, GaussianNoise # pylint: disable=wrong-import-position,no-name-in-module
from tensorflow.keras.models import Model # pylint: disable=wrong-import-position,no-name-in-module
from tensorflow.keras.callbacks import EarlyStopping # pylint: disable=wrong-import-position,no-name-in-module

from ..base import OverSampling, coalesce, coalesce_dict # pylint: disable=wrong-import-position
from ._smote import SMOTE # pylint: disable=wrong-import-position

from .._logger import logger # pylint: disable=wrong-import-position
_logger= logger

__all__= ['DEAGO']

class DEAGO(OverSampling):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{deago,
                            author={Bellinger, C. and Japkowicz, N. and
                                        Drummond, C.},
                            booktitle={2015 IEEE 14th International
                                        Conference on Machine Learning
                                        and Applications (ICMLA)},
                            title={Synthetic Oversampling for Advanced
                                        Radioactive Threat Detection},
                            year={2015},
                            volume={},
                            number={},
                            pages={948-953},
                            keywords={radioactive waste;advanced radioactive
                                        threat detection;gamma-ray spectral
                                        classification;industrial nuclear
                                        facilities;Health Canadas national
                                        monitoring networks;Vancouver 2010;
                                        Isotopes;Training;Monitoring;
                                        Gamma-rays;Machine learning algorithms;
                                        Security;Neural networks;machine
                                        learning;classification;class
                                        imbalance;synthetic oversampling;
                                        artificial neural networks;
                                        autoencoders;gamma-ray spectra},
                            doi={10.1109/ICMLA.2015.58},
                            ISSN={},
                            month={Dec}}

    Notes:
        * There is no hint on the activation functions and amounts of noise.
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_density_estimation,
                  OverSampling.cat_application,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 e=100, # pylint: disable=invalid-name
                 h=0.3, # pylint: disable=invalid-name
                 sigma=0.1,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            e (int): number of epochs
            h (float): fraction of number of hidden units
            sigma (float): training noise
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}

        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0.0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater(e, "e", 1)
        self.check_greater(h, "h", 0)
        self.check_greater(sigma, "sigma", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.ss_params = coalesce_dict(ss_params, ss_params_default)
        self.encoder_params = {'e': e, 'h': h, 'sigma': sigma}
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'e': [40],
                                  'h': [0.1, 0.2, 0.3, 0.4, 0.5],
                                  'sigma': [0.05, 0.1, 0.2]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def execute_autoencoder(self,
                            X_min_normalized, # pylint: disable=invalid-name
                            X_init_normalized # pylint: disable=invalid-name
                            ):
        """
        Execute the autoencoder.

        Args:
            X_min_normalized (np.array): the normalized minority samples
            X_init_normalized (np.array): the normalized generated samples

        Returns:
            np.array: the denoised samples (all)
        """

        n_dim = X_min_normalized.shape[1]
        encoding_d = np.max([2, int(np.rint(n_dim*self.encoder_params['h']))])

        _logger.info("%s: Input dimension: %d, encoding dimension: %d",
                    self.__class__.__name__, n_dim, encoding_d)

        # constructing the autoencoder
        callbacks = [EarlyStopping(monitor='val_loss', patience=2)]

        input_layer = Input(shape=(n_dim,))
        noise = GaussianNoise(self.encoder_params['sigma'])(input_layer)
        encoded = Dense(encoding_d, activation='relu')(noise)
        decoded = Dense(n_dim, activation='linear')(encoded)

        dae = Model(input_layer, decoded)
        dae.compile(optimizer='adadelta', loss='mean_squared_error')
        actual_epochs = np.max([self.encoder_params['e'],
                                int(5000.0/X_min_normalized.shape[0])])

        if X_min_normalized.shape[0] > 10:
            val_num = int(0.2 * X_min_normalized.shape[0])
            X_min_train = X_min_normalized[:-val_num] # pylint: disable=invalid-name
            X_min_val = X_min_normalized[-val_num:] # pylint: disable=invalid-name

            dae.fit(X_min_train,
                    X_min_train,
                    epochs=actual_epochs,
                    validation_data=(X_min_val, X_min_val),
                    callbacks=callbacks,
                    verbose=0)
        else:
            dae.fit(X_min_normalized, X_min_normalized,
                    epochs=actual_epochs, verbose=0)

        return dae.predict(X_init_normalized)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed.")

        tensorflow.random.set_seed(self._random_state_init)

        # sampling by smote
        X_samp, _ = SMOTE(proportion=self.proportion,
                            n_neighbors=self.n_neighbors,
                            nn_params=self.nn_params,
                            ss_params=self.ss_params,
                            n_jobs=self.n_jobs,
                            random_state=self._random_state_init).sample(X, y)

        # samples to map to the manifold extracted by the autoencoder
        X_init = X_samp[len(X):] # pylint: disable=invalid-name

        # normalizing
        X_min = X[y == self.min_label]
        scaler = StandardScaler()
        X_min_normalized = scaler.fit_transform(X_min) # pylint: disable=invalid-name
        X_init_normalized = scaler.transform(X_init) # pylint: disable=invalid-name

        # mapping the initial samples to the manifold
        samples = scaler.inverse_transform(self.execute_autoencoder(X_min_normalized,
                                                                    X_init_normalized))

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'e': self.encoder_params['e'],
                'h': self.encoder_params['h'],
                'sigma': self.encoder_params['sigma'],
                'ss_params': self.ss_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the UnderSampling base class.
"""

from ._base import StatisticsMixin, ParametersMixin, RandomStateMixin

__all__= ['UnderSampling']

class UnderSampling(StatisticsMixin,
                    ParametersMixin,
                    RandomStateMixin):
    """
    Base class of undersampling approaches.
    """

    def __init__(self, random_state=None):
        """
        Constructor
        """
        StatisticsMixin.__init__(self)
        ParametersMixin.__init__(self)
        RandomStateMixin.__init__(self, random_state=random_state)

    def sample(self, X, y):
        """
        Carry out undersampling
        Args:
            X (np.array): features
            y (np.array): labels
        Returns:
            np.array, np.array: sampled X and y
        """
        _, _ = X, y
        return None, None

    def get_params(self, deep=False):
        """
        Returns the parameters of the object as a dictionary.
        Returns:
            dict: the parameters of the object
        """
        return {**RandomStateMixin.get_params(self, deep)}

    def descriptor(self):
        """
        Returns:
            str: JSON description of the current sampling object
        """
        return str((self.__class__.__name__, str(self.get_params())))



"""
This module implements the NRAS method.
"""

import numpy as np

from sklearn.preprocessing import MinMaxScaler
from sklearn.linear_model import LogisticRegression

from ..base import coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['NRAS']

class NRAS(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @article{nras,
                        title = "Noise Reduction A Priori Synthetic
                                    Over-Sampling for class imbalanced data
                                    sets",
                        journal = "Information Sciences",
                        volume = "408",
                        pages = "146 - 161",
                        year = "2017",
                        issn = "0020-0255",
                        doi = "https://doi.org/10.1016/j.ins.2017.04.046",
                        author = "William A. Rivera",
                        keywords = "NRAS, SMOTE, OUPS, Class imbalance,
                                        Classification"
                        }
    """

    categories = [OverSamplingSimplex.cat_sample_ordinary,
                  OverSamplingSimplex.cat_noise_removal,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 t=0.5,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            n_neighbors (int): number of neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            t (float): [0,1] fraction of n_neighbors as threshold
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_in_range(t, "t", [0, 1])
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.t = t # pylint: disable=invalid-name
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [5, 7, 9],
                                  't': [0.3, 0.5, 0.8]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def propensity_scores(self,
                            X_trans, # pylint: disable=invalid-name
                            y):
        """
        Determining the propensity scores.

        Args:
            X_trans (np.array): normalized training vectors
            y (np.array): the target labels

        Returns:
            np.array: the propensity scores
        """
        # determining propensity scores using logistic regression
        logreg = LogisticRegression(solver='lbfgs',
                                n_jobs=self.n_jobs,
                                random_state=self._random_state_init)
        logreg.fit(X_trans, y)
        propensity = logreg.predict_proba(X_trans)
        propensity = propensity[:, np.where(logreg.classes_ == self.min_label)[0][0]]

        return propensity

    def neighborhood_structure(self,
                                X_new,
                                y,
                                n_neighbors,
                                X_min_new # pylint: disable=invalid-name
                                ):
        """
        Determine neighborhood structure.

        Args:
            X_new (np.array): the extended training vectors
            y (np.array): the target labels
            n_neighbors (int): the number of neighbors
            X_min_new (np.array): the minority samples

        Returns:
            np.array: the neighborhood structure
        """
        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
                    self.metric_tensor_from_nn_params(nn_params, X_new, y)

        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_new)
        ind = nnmt.kneighbors(X_min_new, return_distance=False)

        return ind

    def generate_samples(self, *, X_min, to_remove, X_trans, y,
                                                    ind, n_to_sample):
        """
        Generate samples

        Args:
            X_min (np.array): the minority samples
            to_remove (np.array): minority indices flagged as noise
            X_trans (np.array): all training samples
            y (np.array): all target labels
            ind (np.array): the neighborhood structure
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """
        base_weights = np.repeat(1.0, len(X_min))
        base_weights[to_remove] = 0.0
        vertex_weights = np.repeat(1.0, len(X_trans))
        vertex_weights[y == self.min_label][to_remove] = 0.0

        samples = self.sample_simplex(X=X_min,
                                        indices=ind,
                                        n_to_sample=n_to_sample,
                                        X_vertices=X_trans,
                                        vertex_weights=vertex_weights,
                                        base_weights=base_weights)
        return samples

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed.")

        # standardization is needed to make the range of the propensity scores
        # similar to that of the features
        mms = MinMaxScaler()
        X_trans = mms.fit_transform(X) # pylint: disable=invalid-name

        X_min = X_trans[y == self.min_label]

        # adding propensity scores as a new feature
        X_new = np.column_stack([X_trans, self.propensity_scores(X_trans, y)])
        X_min_new = X_new[y == self.min_label] # pylint: disable=invalid-name

        # finding nearest neighbors of minority samples
        n_neighbors = min([len(X_new), self.n_neighbors+1])

        ind = self.neighborhood_structure(X_new, y, n_neighbors, X_min_new)

        # noise removal
        t_hat = np.sum(y[ind[:, 1:]] == self.min_label, axis=1)
        to_remove = np.where(t_hat < self.t * n_neighbors)[0]

        if len(to_remove) >= len(X_min) - 1:
            return self.return_copies(X, y,
                            "most minority samples indentified as noise")

        n_to_sample = n_to_sample + to_remove.shape[0]

        samples = self.generate_samples(X_min=X_min,
                                        to_remove=to_remove,
                                        X_trans=X_trans,
                                        y=y,
                                        ind=ind,
                                        n_to_sample=n_to_sample)

        X_min = np.delete(X_min, to_remove, axis=0)

        # do the sampling
        #samples = []
        #while len(samples) < n_to_sample:
        #    idx = self.random_state.randint(len(X_min))
        #    # finding the number of minority neighbors
        #    t_hat = np.sum(y[ind[idx][1:]] == self.min_label)
        #    if t_hat < self.t*n_neighbors:
        #        # removing the minority point if the number of minority
        #        # neighbors is less then the threshold
        #        # to_remove indexes X_min
        #        if idx not in to_remove:
        #            to_remove.append(idx)
        #            # compensating the removal of the minority point
        #            n_to_sample = n_to_sample + 1
        #
        #        if len(to_remove) == len(X_min):
        #            _logger.warning(self.__class__.__name__ + ": " +
        #                            "all minority samples identified as noise")
        #            return X.copy(), y.copy()
        #    else:
        #        # otherwise do the sampling
        #        X_b = X_trans[self.random_state.choice(ind[idx][1:])]
        #        samples.append(self.sample_between_points(X_min[idx], X_b))

        return (mms.inverse_transform(np.vstack([X_trans[y == self.maj_label],
                                                 X_min,
                                                 samples])),
                np.hstack([np.repeat(self.maj_label,
                                                np.sum(y == self.maj_label)),
                           np.repeat(self.min_label, len(X_min)),
                           np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                't': self.t,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the OUPS method.
"""

import numpy as np

from sklearn.linear_model import LogisticRegression

from ..base import coalesce_dict
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['OUPS']

class OUPS(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @article{oups,
                        title = "A priori synthetic over-sampling methods for
                                    increasing classification sensitivity in
                                    imbalanced data sets",
                        journal = "Expert Systems with Applications",
                        volume = "66",
                        pages = "124 - 135",
                        year = "2016",
                        issn = "0957-4174",
                        doi = "https://doi.org/10.1016/j.eswa.2016.09.010",
                        author = "William A. Rivera and Petros Xanthopoulos",
                        keywords = "SMOTE, OUPS, Class imbalance,
                                    Classification"
                        }

    Notes:
        * In the description of the algorithm a fractional number p (j) is
            used to index a vector.
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_sample_ordinary]

    def __init__(self,
                 proportion=1.0,
                 *,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)

        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def determine_sorting(self, X, y):
        """
        Determine the sorting of vectors.

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels

        Returns:
            np.array: the sorting
        """
        # extracting propensity scores
        logreg = LogisticRegression(solver='lbfgs',
                                n_jobs=self.n_jobs,
                                random_state=self._random_state_init)
        logreg.fit(X, y)
        propensity = logreg.predict_proba(X)
        propensity = propensity[:,
                    np.where(logreg.classes_ == self.min_label)[0][0]]

        sorting = propensity.argsort()[::-1]

        return sorting

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels
        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        sorting = self.determine_sorting(X, y)

        X_sorted = X[sorting] # pylint: disable=invalid-name

        min_indices = y[sorting] == self.min_label
        X_sorted_min = X_sorted[min_indices] # pylint: disable=invalid-name
        p = np.sum(y == self.maj_label) / np.sum(y == self.min_label) # pylint: disable=invalid-name
        simplex_weights = np.tile(np.arange(p, 0, -1.0), (X.shape[0], 1))
        simplex_weights = np.clip(simplex_weights, 0.0, 1.0)

        indices = np.vstack([np.arange(0, len(X)) + idx
                            for idx in range(1, simplex_weights.shape[1] + 1)]).T

        simplex_weights[indices >= len(X)] = 0.0
        indices = indices[min_indices]

        mask = ~ np.all(indices >= X.shape[0], axis=1)

        indices = indices[mask]
        X_sorted_min = X_sorted_min[mask] # pylint: disable=invalid-name
        indices[indices >= X.shape[0]] = X.shape[0] - indices[indices >= X.shape[0]] - 1

        #n_dim_orig = self.n_dim
        #self.n_dim = np.min([self.n_dim, simplex_weights.shape[1] + 1])

        samples = self.sample_simplex(X=X_sorted_min,
                                        indices=indices,
                                        n_to_sample=n_to_sample,
                                        X_vertices=X_sorted,
                                        simplex_weights=simplex_weights)
        #self.n_dim = n_dim_orig

        # sorting indices according to propensity scores
        #prop_sorted = sorted(zip(propensity,
        #                np.arange(len(propensity))), key=lambda x: -x[0])
        #p = np.sum(y == self.maj_label) / np.sum(y == self.min_label)
        #n = 0
        #samples = []
        #
        ## implementing Algorithm 1 in the cited paper with some minor changes
        ## to enable the proper sampling of p numbers
        #while n < len(propensity) and len(samples) < n_to_sample:
        #    if (y[prop_sorted[n][1]] == self.min_label
        #            and n < len(propensity) - 1):
        #        num = 1
        #        p_tmp = p
        #        while p_tmp > 0 and n + num < len(propensity):
        #            if self.random_state.random_sample() < p_tmp:
        #                samples.append(self.sample_between_points(
        #                    X[prop_sorted[n][1]], X[prop_sorted[n+num][1]]))
        #            p_tmp = p_tmp - 1
        #            num = num + 1
        #    n = n + 1

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the Gazzah method.
"""

import numpy as np

from sklearn.decomposition import PCA

from ..base import OverSampling
from ._polynom_fit_smote import polynom_fit_SMOTE_star

from .._logger import logger
_logger= logger

__all__= ['Gazzah']

class Gazzah(OverSampling):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{gazzah,
                            author={Gazzah, S. and Hechkel, A. and Essoukri
                                        Ben Amara, N. },
                            booktitle={2015 IEEE 12th International
                                        Multi-Conference on Systems,
                                        Signals Devices (SSD15)},
                            title={A hybrid sampling method for
                                    imbalanced data},
                            year={2015},
                            volume={},
                            number={},
                            pages={1-6},
                            keywords={computer vision;image classification;
                                        learning (artificial intelligence);
                                        sampling methods;hybrid sampling
                                        method;imbalanced data;
                                        diversification;computer vision
                                        domain;classical machine learning
                                        systems;intraclass variations;
                                        system performances;classification
                                        accuracy;imbalanced training data;
                                        training data set;over-sampling;
                                        minority class;SMOTE star topology;
                                        feature vector deletion;intra-class
                                        variations;distribution criterion;
                                        biometric data;true positive rate;
                                        Training data;Principal component
                                        analysis;Databases;Support vector
                                        machines;Training;Feature extraction;
                                        Correlation;Imbalanced data sets;
                                        Intra-class variations;Data analysis;
                                        Principal component analysis;
                                        One-against-all SVM},
                            doi={10.1109/SSD.2015.7348093},
                            ISSN={},
                            month={March}}
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_dim_reduction,
                  OverSampling.cat_changes_majority]

    def __init__(self,
                 proportion=1.0,
                 *,
                 n_components=2,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_components (int): number of components in PCA analysis
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_components, "n_components", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_components = n_components
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_components': [2, 3, 4, 5]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        self.class_label_statistics(y)

        # do the oversampling
        pf_smote = polynom_fit_SMOTE_star(proportion=self.proportion,
                                     random_state=self._random_state_init)
        X_samp, _ = pf_smote.sample(X, y)
        X_min_samp = X_samp[len(X):] # pylint: disable=invalid-name

        if len(X_min_samp) == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        # do the undersampling
        X_maj = X[y == self.maj_label]

        # fitting the PCA model
        pca = PCA(n_components=min([len(X[0]), self.n_components]))
        X_maj_trans = pca.fit_transform(X_maj) # pylint: disable=invalid-name
        R = np.sqrt(np.sum(np.var(X_maj_trans, axis=0))) # pylint: disable=invalid-name

        # determining the majority samples to remove
        to_remove = np.where(np.linalg.norm(X_maj_trans, axis=1) > R)[0]

        _logger.info("%s: Removing %d majority samples",
                        self.__class__.__name__, len(to_remove))

        # removing the majority samples
        X_maj = np.delete(X_maj, to_remove, axis=0)

        return (np.vstack([X_maj, X_min_samp]),
                np.hstack([np.repeat(self.maj_label, len(X_maj)),
                           np.repeat(self.min_label, len(X_min_samp))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_components': self.n_components,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the DSMOTE method.
"""
from dataclasses import dataclass

import numpy as np

from sklearn.preprocessing import MinMaxScaler
from scipy.stats.mstats import gmean

from ..base import (NearestNeighborsWithMetricTensor,
                                pairwise_distances_mahalanobis)
from ..base import OverSampling

from .._logger import logger
_logger= logger

__all__= ['DSMOTE']

@dataclass
class DSMOTEPartialResults:
    """
    This class is used to pass around some partial calculation results
    reused at various parts of the algorithm.

    Args:
        log_sum (np.array): the array of sums of logarithms of minority
                                data coordinate-wise
        sum1 (np.array): the array of the sums of minority coordinates
        sum2 (np.array): the array of the sums of squares of minority coordinates
        all_sum (float): the sum of all minority coordinate values
        norm (np.array): the norms of all minority samples
    """
    log_sum: np.array
    sum1: np.array
    sum2: np.array
    all_sum: float
    norm: np.array

class DSMOTE(OverSampling):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{dsmote,
                            author={Mahmoudi, S. and Moradi, P. and Akhlaghian,
                                    F. and Moradi, R.},
                            booktitle={2014 4th International Conference on
                                        Computer and Knowledge Engineering
                                        (ICCKE)},
                            title={Diversity and separable metrics in
                                    over-sampling technique for imbalanced
                                    data classification},
                            year={2014},
                            volume={},
                            number={},
                            pages={152-158},
                            keywords={learning (artificial intelligence);
                                        pattern classification;sampling
                                        methods;diversity metric;separable
                                        metric;over-sampling technique;
                                        imbalanced data classification;
                                        class distribution techniques;
                                        under-sampling technique;DSMOTE method;
                                        imbalanced learning problem;diversity
                                        measure;separable measure;Iran
                                        University of Medical Science;UCI
                                        dataset;Accuracy;Classification
                                        algorithms;Vectors;Educational
                                        institutions;Euclidean distance;
                                        Data mining;Diversity measure;
                                        Separable Measure;Over-Sampling;
                                        Imbalanced Data;Classification
                                        problems},
                            doi={10.1109/ICCKE.2014.6993409},
                            ISSN={},
                            month={Oct}}

    Notes:
        * The method is highly inefficient when the number of minority samples
            is high, time complexity is O(n^3), with 1000 minority samples it
            takes about 1e9 objective function evaluations to find 1 new sample
            points. Adding 1000 samples would take about 1e12 evaluations of
            the objective function, which is unfeasible. We introduce a new
            parameter, n_step, and during the search for the new sample at
            most n_step combinations of minority samples are tried.
        * Abnormality of minority points is defined in the paper as
            D_maj/D_min, high abnormality  means that the minority point is
            close to other minority points and very far from majority points.
            This is definitely not abnormality,
            I have implemented the opposite.
        * Nothing ensures that the fisher statistics and the variance from
            the geometric mean remain comparable, which might skew the
            optimization towards one of the sub-objectives.
        * MinMax normalization doesn't work, each attribute will have a 0
            value, which will make the geometric mean of all attribute 0.
    """

    categories = [OverSampling.cat_changes_majority,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params={},
                 rate=0.1,
                 n_step=50,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors in nearest neighbors
                                component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            rate (float): [0,1] rate of minority samples to turn into majority
            n_step (int): number of random configurations to check for new
                                samples
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_in_range(rate, "rate", [0, 1])
        self.check_greater_or_equal(n_step, "n_step", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = nn_params
        self.rate = rate
        self.n_step = n_step
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'rate': [0.1, 0.2],
                                  'n_step': [50]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def calculate_gdiv(self,
                        X_min_new, # pylint: disable=invalid-name
                        gmean_min,
                        partial):
        """
        Calculate the g-divergence score.

        Args:
            X_min_new (np.array): the new minority vectors
            gmean_min (np.array): the geometric mean of the minority points
            partial (DSMOTEPartialResults): partial results

        Returns:
            float: the g-divergence
        """
        inner_prod = np.dot(X_min_new, gmean_min)
        gmean_norm = np.linalg.norm(gmean_min)**2
        term_sum = partial.norm - 2*inner_prod + gmean_norm
        gdiv = np.mean(np.sqrt(term_sum))

        return gdiv

    def calculate_fisher(self, all_mean_min, mean_maj, var_min, var_maj):
        """
        Calculate the Fisher ratio

        Args:
            all_mean_min (float): mean of all minority vectors
            mean_maj (float): mean of the majority vectors
            var_min (float): mean coordinate-wise variance of all
                                minority vectors
            var_maj (float): mean coordinate-wise variances of
                                majority vectors

        Returns:
            float: the Fisher ratio
        """
        fisher_numerator = (all_mean_min - mean_maj)**2
        fisher_denominator = np.mean(var_min) + var_maj
        fisher = fisher_numerator / fisher_denominator
        return fisher

    def calculate_objective(self,
                            X_min_new, # pylint: disable=invalid-name
                            partial,
                            mean_maj,
                            var_maj):
        """
        Calculate the objective function.

        Args:
            X_min_new (np.array): new minority samples
            partial (DSMOTEPartialResults): partial results
            mean_maj (float): mean of the majority vectors
            var_maj (float): mean coordinate-wise variance of the
                                majority vectors

        Returns:
            float: the objective function value
        """
        # computing mean, var, gmean and mean of all elements with
        # the new sample (gm)
        mean_min = partial.sum1 / X_min_new.shape[0]
        var_min = partial.sum2 / X_min_new.shape[0] - mean_min**2
        gmean_min = np.exp(partial.log_sum / X_min_new.shape[0])
        all_n_min = np.prod(X_min_new.shape)
        all_mean_min = partial.all_sum / all_n_min

        # computing the new objective function value
        score = self.calculate_gdiv(X_min_new, gmean_min, partial) \
                + self.calculate_fisher(all_mean_min, mean_maj, var_min, var_maj)

        return score

    def generate_one_sample(self, X_min, mean_maj, var_maj, partial_results):
        """
        Generates one new sample

        Args:
            X_min (np.array): minority vectors
            mean_maj (float): mean of the majority vectors
            var_maj (float): mean coordinate-wise variance of the
                                majority vectors
            partial_results (DSMOTEPartialResults): partial calculation results

        Returns:
            np.array, DSMOTEPartialResults: the new vector and the
                                            partial results
        """
        best_candidate = None
        new_partial_results = {}
        highest_score = 0.0
        # we try n_step combinations of minority samples
        n_steps = np.min([X_min.shape[0] \
                        * (X_min.shape[0]-1) \
                        * (X_min.shape[0]-2), self.n_step])

        for _ in range(n_steps):
            indices = self.random_state.choice(np.arange(X_min.shape[0]),
                                                        3,
                                                        replace=False)
            gmv = gmean(X_min[indices], axis=0)

            # computing the new objective function for the new point (gm)
            #  added
            X_min_new = np.vstack([X_min, gmv]) # pylint: disable=invalid-name

            # updating the components of the objective function
            partial_tmp = DSMOTEPartialResults(
                            log_sum=partial_results.log_sum + np.log(gmv),
                            sum1=partial_results.sum1 + gmv,
                            sum2=partial_results.sum2 + gmv**2,
                            all_sum=partial_results.all_sum + np.sum(gmv),
                            norm=partial_results.norm + np.linalg.norm(gmv))

            score = self.calculate_objective(X_min_new, partial_tmp, mean_maj, var_maj)
            # evaluate the objective function
            # score= objective(np.vstack([X_min, gm]))
            # check if the score is better than the best so far
            if score > highest_score:
                highest_score = score
                best_candidate = gmv
                new_partial_results = partial_tmp

        return best_candidate, new_partial_results


    def sampling(self, X_maj, X_min, n_to_sample):
        """
        Carry out the sampling

        Args:
            X_maj (np.array): the majority samples
            X_min (np.array): the minority samples
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the updated minority samples
        """
        # computing the mean and variance of points in the majority class
        var_maj = np.mean(np.var(X_maj, axis=0))
        mean_maj = np.mean(X_maj)

        # records the sum of logarithms in X_min, used to compute the geometric
        # mean
        log_sum = np.sum(np.log(X_min), axis=0)
        # contains the sum of values in X_min, coordinatewise
        sum1 = np.sum(X_min, axis=0)
        # contains the squares of sums of values in X_min, coordinatewise
        sum2 = np.sum(X_min**2, axis=0)
        # contains the sum of all numbers in X_min
        all_sum = np.sum(X_min)

        norm = np.linalg.norm(X_min)**2

        partial_results = DSMOTEPartialResults(log_sum=log_sum,
                                                sum1=sum1,
                                                sum2=sum2,
                                                all_sum=all_sum,
                                                norm=norm)

        # do the sampling
        n_added = 0
        while n_added < n_to_sample:
            best_candidate, partial_results = self.generate_one_sample(X_min,
                                                                mean_maj,
                                                                var_maj,
                                                                partial_results)
            # add the best candidate to the minority samples
            X_min = np.vstack([X_min, best_candidate])
            n_added = n_added + 1

        return X_min

    def calculate_abnormality(self, X, y, X_min, X_maj):
        """
        Calculate the abnormality scores.

        Args:
            X (np.array): all features
            y (np.array): all labels
            X_min (np.array): minority samples
            X_maj (np.array): majority samples

        Returns:
            np.array: the abnormality scores
        """
        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        # fitting nearest neighbors model
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=len(X_maj),
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_maj)
        dist, _ = nnmt.kneighbors(X_min)

        # compute mean distances, the D_min is compenstaed for taking into
        # consideration self-distances in the mean
        D_maj = np.mean(dist, axis=1) # pylint: disable=invalid-name
        D_min = np.mean(pairwise_distances_mahalanobis(X_min, # pylint: disable=invalid-name
                                tensor=nn_params.get('metric_tensor', None)), axis=1)
        D_min = D_min * len(X_min)/(len(X_min)-1) # pylint: disable=invalid-name

        # computing degree of abnormality
        abnormality = D_min/D_maj

        return abnormality

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        mms = MinMaxScaler(feature_range=(1e-6, 1.0 - 1e-6))
        X = mms.fit_transform(X)

        X_min = X[y == self.min_label]
        X_maj = X[y == self.maj_label]

        abnormality = self.calculate_abnormality(X, y, X_min, X_maj)

        # sorting minority indices in decreasing order by abnormality
        to_sort = zip(abnormality, np.arange(len(abnormality)))
        abnormality, indices = zip(*sorted(to_sort, key=lambda x: -x[0]))
        rate = int(self.rate*len(abnormality))

        if rate > 0:
            # moving the most abnormal points to the majority class
            X_maj = np.vstack([X_maj, X_min[np.array(indices[:rate])]])
            # removing the most abnormal points form the minority class
            X_min = np.delete(X_min, indices[:rate], axis=0)

        # this is the original objective function, however, using this
        # is very inefficient if the number of records increases above
        # approximately 1000
        # def objective(X):
        #    """
        #    The objective function to be maximized
        #
        #    Args:
        #        X (np.array): dataset
        #
        #    Returns:
        #        float: the value of the objective function
        #    """
        #    gm= gmean(X, axis= 0)
        #    gdiv= np.mean(np.linalg.norm(X - gm, axis= 1))
        #    fisher= (np.mean(X) - mean_maj)**2/(np.mean(np.var(X, axis= 0)) \
        #                + var_maj)
        #    return gdiv + fisher

        # in order to make the code more efficient, we do maintain some
        # variables containing the main componentes of the objective function
        # and apply only small corrections based on the new values being added
        # the effect should be identical

        X_min = self.sampling(X_maj, X_min, n_to_sample)

        return (mms.inverse_transform(np.vstack([X_maj, X_min])),
                np.hstack([np.repeat(self.maj_label, len(X_maj)),
                           np.repeat(self.min_label, len(X_min))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'rate': self.rate,
                'n_step': self.n_step,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}




"""
This module implements all simplex sampling related functionalities.
"""

import itertools

import numpy as np

from ._base import RandomStateMixin, coalesce
from .._logger import logger as _logger

__all__= ['array_array_index',
          'base_idx_neighbor_idx_simplices',
          'all_neighbor_simplices_real_idx',
          'reweight_simplex_vertices',
          'cartesian_product',
          'vector_choice',
          'simplex_volume',
          'simplex_volumes',
          'SimplexSamplingMixin',
          'random_samples_from_simplices',
          'add_samples']

def array_array_index(array, indices):
    """
    Take samples from an array by indices row by row

    Args:
        array (np.array): array to sample from
        indices (np.array): array of indices

    Returns:
        array (indices.shape): the sampled values
    """
    stride = np.arange(indices.shape[0])*array.shape[1]
    indices_mod = indices + stride[:, None]
    indices_flat = indices_mod.flatten()
    return array.flatten()[indices_flat].reshape(indices.shape)

def base_idx_neighbor_idx_simplices(n_base, n_neighbors=5, n_dim=2):
    """
    Returns all possible simplices of a given dimensionality

    Args:
        n_base (int): number of base records
        n_neighbors (int): number of neighbors
        n_dim (int): simplex dimensionality (for triangles 3)

    Returns:
        np.array: first column: base record index, the rest are neighbor
                                indices
    """
    combinations = np.array(list(itertools.combinations(np.arange(1,
                                                        n_neighbors),
                                                        n_dim-1))).astype(int)
    base_indices = np.repeat(np.arange(n_base), len(combinations))
    all_simplices = np.vstack([base_indices,
                                np.tile(combinations, (n_base, 1)).T]).T
    return all_simplices

def all_neighbor_simplices_real_idx(n_dim, indices):
    """
    Determine the real indices of all simplices

    Args:
        n_dim (int): simplex order
        indices (np.array): neighbor indices

    Returns:
        np.array: the array of simplex indices of the real vectors
    """
    all_simplices = base_idx_neighbor_idx_simplices(n_base=indices.shape[0],
                                                n_neighbors=indices.shape[1],
                                                n_dim=n_dim)
    base_vector_indices = all_simplices[:, 0]
    neighbors_indices = indices[base_vector_indices]
    neighbors_indices = array_array_index(neighbors_indices,
                                            all_simplices[:,1:])
    simplices_real_indices = np.vstack([base_vector_indices.T,
                                        neighbors_indices.T]).T

    return simplices_real_indices

def simplex_volume(simplex):
    """
    Computes the volume of a simplex

    Args:
        simplex (np.array): vertices of the simplex

    Returns:
        float: the volume of the simplex
    """
    simplex_mod = simplex[:-1] - simplex[-1]
    gram = np.dot(simplex_mod, simplex_mod.T)
    det = np.linalg.det(gram)
    return np.sqrt(det)/np.math.factorial(simplex.shape[0]-1)

def simplex_volumes(simplices):
    """
    Computes the volume of an array of simplices

    Args:
        simplices (np.array): array of the simplices' vertices

    Returns:
        np.array: the volumes of the simplices
    """
    if simplices.shape[0] == 0:
        return np.array([], dtype=float)
    if simplices[0].shape[0] == 2:
        return np.sqrt(np.sum((simplices[:,0,:] - simplices[:,1,:])**2,
                                axis=1))
    return np.array([simplex_volume(x) for x in simplices])

def reweight_simplex_vertices(X,
                                simplices,
                                X_vertices=None,
                                vertex_weights=None):
    """
    Simplex vertices are split into two sets, and the weights of the ones in
    the set X_vertices vary. The weight 1 means full contribution, a weight
    0 < w < 1 a proportionally smaller contribution.

    Args:
        X (np.array): the base vectors
        simplices (np.array): the vertices of the simplices
        X_vertices (np.array): the neighbor vectors
        vertex_weights (np.array): the weights of the neighbor vectors

    Returns:
        np.array: the simplices with scaled neighbor edges
    """
    if X_vertices is None:
        X_vertices = X

    # extracting the weights of each neighbor vector
    vertex_weights_simplices = vertex_weights[simplices[:,1:]]

    X_base = X[simplices[:,0]]
    X_neighbors = X_vertices[simplices[:,1:]]

    # calculating the direction vectors (neighbor vectors - base vector)
    diff_vectors = X_neighbors - X_base[:,None]

    # reweighting the direction vectors of neighbors
    weighted = (diff_vectors.T*vertex_weights_simplices.T).T

    # shifint the scaled neighbors back by the base vectors
    reverted = weighted + X_base[:,None]

    # concatenating the base vectors and the scaled vectors
    return np.concatenate([X_base[:,None], reverted], axis=1)

def cartesian_product(*arrays):
    """
    Computes the Cartesian product of arrays

    Args:
        arrays: variable list of arrays

    Returns:
        array(np.array): the array of Cartesian products
    """
    length = len(arrays)
    dtype = np.result_type(*arrays)
    arr = np.empty([len(a) for a in arrays] + [length], dtype=dtype)
    for idx, array in enumerate(np.ix_(*arrays)):
        arr[...,idx] = array
    return arr.reshape(-1, length)

def vector_choice(arr,
                    p, # pylint: disable=invalid-name
                    random_state=None):
    """
    Vector choice

    Args:
        arr (np.array): the array to choose from
        p (np.array): the array of probabilities

    Returns:
        np.array: the choosen elements
    """
    if random_state is None:
        random_state = np.random.RandomState(5)

    prob_cum = np.cumsum(p, axis=1)
    rands = random_state.rand(p.shape[0])
    return arr[np.argmax((rands < prob_cum.T).T, axis=1)]

def random_samples_from_simplices(X,
                                    simplices,
                                    X_vertices=None,
                                    vertex_weights=None,
                                    random_state=None):
    """
    Generate random samples within the simplices.

    Args:
        X (np.array): the base vectors
        simplices (np.array): the indices of the simplex nodes
        X_vertices (np.array): possibly different array of neighboring
                                vertices
        vertex_weights (np.array): weights of the neighboring vertices
        random_state (np.random.RandomState/None): the random state to
                                                    use

    Returns:
        np.array: the random samples
    """
    if random_state is None:
        random_state = np.random.RandomState(5)

    weights = random_state.random_sample(simplices.shape)
    weights = weights/np.sum(weights, axis=1)[:, None]

    if X_vertices is None and vertex_weights is None:
        X_simp = X[simplices]
    elif X_vertices is None and vertex_weights is not None:
        X_simp = reweight_simplex_vertices(X,
                                            simplices,
                                            X,
                                            vertex_weights)
    elif X_vertices is not None and vertex_weights is None:
        n_records = X.shape[0]
        X_all = np.vstack([X, X_vertices])
        simplices[:,1:] = simplices[:,1:] + n_records
        X_simp = X_all[simplices]
        simplices[:,1:] = simplices[:,1:] - n_records
    else:
        X_simp = reweight_simplex_vertices(X,
                                        simplices,
                                        X_vertices,
                                        vertex_weights)

    weighted_simplices = X_simp*weights[:, :, None]

    samples = np.sum(weighted_simplices, axis=1)

    return samples

def add_samples(*,
                pairs,
                counts,
                count,
                X,
                X_vertices,
                vertex_weights=None):
    """
    Create samples for 'count' splitting points
    between the edge pairs listed in 'pairs'

    Args:
        pairs (np.array(tuple)): unique simplices
        counts (np.array(int)): cardinalities
        count (int): one particular count
        X (np.array): base points
        X_vertices (np.array): possibly different neighbors
        vertex_weights (np.array): weights of the vertices

    Returns:
        np.array: the generated samples
    """
    pairs_filtered = pairs[counts == count]

    splits = 1.0/(count + 1)*(np.arange(1, count+1))

    base_points = X[pairs_filtered[:,0]]
    neighbor_points = X_vertices[pairs_filtered[:,1]]

    if vertex_weights is None:
        weights = np.repeat(1.0, len(pairs_filtered))
    else:
        weights = vertex_weights[pairs_filtered[:,1]]

    diffs = neighbor_points - base_points
    diffweight = (diffs.T*weights).T

    samples_by_count = [base_points + diffweight*s for s in splits]
    return np.vstack(samples_by_count)

class SimplexSamplingMixin(RandomStateMixin):
    """
    The mixin class for all simplex sampling based techniques.
    """
    def __init__(self,
                *,
                n_dim=2,
                simplex_sampling='uniform',
                within_simplex_sampling='deterministic',
                gaussian_component=None,
                random_state=None):
        """
        Base mixin of the methods implementing simplex sampling.

        Args:
            n_dim (int): the dimensions of the simplices
                            (2 for line segments)
            simplex_sampling (str): simplex sampling method
                                    ('uniform'/'volume'/None)
            within_simplex_sampling (str): within simplex sampling
                                    method ('random'/'deterministic')
        """
        RandomStateMixin.__init__(self, random_state)
        self.n_dim = n_dim
        self.simplex_sampling = simplex_sampling
        self.within_simplex_sampling = within_simplex_sampling
        self.gaussian_component = coalesce(gaussian_component, {})

    def get_params(self, deep=False):
        """
        Returns the params of the object.

        Args:
            deep (bool): deep discovery of params

        Returns:
            dict: the parameter dict
        """
        sampling_params = {'n_dim': self.n_dim,
                    'simplex_sampling': self.simplex_sampling,
                    'within_simplex_sampling': self.within_simplex_sampling,
                    'gaussian_component': self.gaussian_component}
        return {'ss_params': sampling_params,
                **RandomStateMixin.get_params(self, deep)}

    def determine_simplex_distribution(self, X, simplices):
        """
        Determine the simplex distribution.

        Args:
            X (np.array): the base vectors
            simplices (np.array): the node indices of the simplices

        Returns:
            np.array: the distribution for sampling the simplices
        """
        if self.simplex_sampling == 'uniform':
            return np.repeat(1.0/len(simplices), len(simplices))
        if self.simplex_sampling == 'volume':
            return simplex_volumes(X[simplices])
        raise ValueError(f"simplex sampling with weighting"\
                            f"{self.simplex_sampling} not implemented yet")

    def all_simplices_node_weights(self, indices, simplex_weights, n_dim):
        """
        Determine all simplices and the node weights.

        Args:
            indices (np.array): the neighborhood structure
            simplex_weights (np.array): the simplex node weights
            n_dim (int): the dimensionality of the simplices

        Returns:
            np.array, np.array: all simplices and the node weights
        """
        all_simplices = all_neighbor_simplices_real_idx(n_dim, indices)
        if simplex_weights is not None:
            all_simplices_nidx = base_idx_neighbor_idx_simplices(n_base=indices.shape[0],
                                                            n_neighbors=indices.shape[1],
                                                            n_dim=n_dim)
            base_indices = all_simplices_nidx[:,0]
            neighbor_indices = all_simplices_nidx[:,1:]
            node_weights = array_array_index(simplex_weights[base_indices], neighbor_indices)
            node_weights = np.prod(node_weights, axis=1)
        else:
            node_weights = np.repeat(1.0, all_simplices.shape[0])

        return all_simplices, node_weights

    def determine_weights(self, *, X, base_weights, X_vertices, all_simplices, indices):
        """
        Determine the weights.

        Args:
            X (np.array): the base vectors
            base_weights (np.array): the base weights
            X_vertices (np.array): the vertices
            all_simplices (np.array): all simplices
            indices (np.array): the neighborhood structure

        Returns:
            np.array: the weights
        """
        if base_weights is None:
            # base_weights overwrites the simplex_sampling parameter
            if X_vertices is None:
                weights = self.determine_simplex_distribution(X, all_simplices)
            else:
                # joining all vectors to make things easier
                simplices_joint = all_simplices.copy()
                simplices_joint[:, 1:] = simplices_joint[:, 1:]\
                                        + indices.shape[0]
                X_joint = np.vstack([X, X_vertices])
                weights = self.determine_simplex_distribution(X_joint,
                                                            simplices_joint)
        else:
            n_records = len(base_weights)
            n_simplices = len(all_simplices)
            weights = np.repeat(base_weights, int(n_simplices/n_records))

        return weights

    def simplices(self,
                    X,
                    n_to_sample,
                    *,
                    base_weights=None,
                    indices=None,
                    X_vertices=None,
                    simplex_weights=None,
                    n_dim=None):
        """
        Sampling the simplices.

        Args:
            X (np.array): the base vectors
            n_to_sample (int): the number of samples to generate
            base_weights (np.array): the weights of the base vectors
                        (overwrites the simplex_sampling parameter)
            indices (np.array): the neighborhood relations
            X_vertices (np.array): possibly different array for neighbor
                                    vectors
            simplex_weights (np.array): weights of the simplices
        """
        all_simplices, node_weights = \
            self.all_simplices_node_weights(indices, simplex_weights, n_dim=n_dim)

        weights = self.determine_weights(X=X,
                                         base_weights=base_weights,
                                         X_vertices=X_vertices,
                                         all_simplices=all_simplices,
                                         indices=indices)

        weights = weights * node_weights

        # sample the simplices
        choices = np.arange(all_simplices.shape[0])
        selected_indices = self.random_state.choice(choices,
                                                    n_to_sample,
                                                    p=weights/np.sum(weights))
        return all_simplices[selected_indices]

    def add_gaussian_noise(self, samples):
        """
        Add Gaussian noise according to the specification.

        Args:
            samples (np.array): the generated samples

        Returns:
            np.array: the noisy samples
        """

        if 'sigma' in self.gaussian_component:
            sigma = self.gaussian_component['sigma']
            return samples + self.random_state.normal(size=samples.shape) * sigma
        if 'sigmas' in self.gaussian_component:
            sigmas = self.gaussian_component['sigmas']
            return samples + self.random_state.normal(size=samples.shape) * sigmas

        return samples

    def deterministic_samples_dim_2(self,
                                    X,
                                    simplices,
                                    X_vertices=None,
                                    vertex_weights=None):
        """
        Carry out uniform sampling on line segments.

        Args:
            X (np.array): the base vectors
            simplices (np.array): the indexes of the simplex nodes
            X_vertices (np.array): array of neighbor vectors
            vertex_weights (np.array): weights of the neighbor vectors

        Returns:
            np.array: the random samples
        """
        samples = []

        base_indices = simplices[:, 0]
        neighbor_indices = simplices[:, 1]

        pairs = np.array(list(zip(base_indices, neighbor_indices)))

        if X_vertices is None or X is X_vertices:
            pairs.sort()
            X_vertices = X

        values, counts = np.unique(pairs, axis=0, return_counts=True)

        max_count = np.max(counts)

        samples = [add_samples(pairs=values,
                                counts=counts,
                                count=count,
                                X=X,
                                X_vertices=X_vertices,
                                vertex_weights=vertex_weights) for count in range(1, max_count+1)\
                                                         if count in counts]
        return np.vstack(samples)

    def deterministic_samples_dim_1(self,
                                    X,
                                    simplices):
        """
        Deterministic samples for 1-st order simplices.

        Args:
            X (np.array): base vectors
            simplices (np.array): the simplices

        Returns:
            np.array: the sampled simlices
        """
        return X[simplices[:, 0]]

    def deterministic_samples_from_simplices(self,
                                                *,
                                                X,
                                                simplices,
                                                X_vertices=None,
                                                vertex_weights=None,
                                                n_dim=None):
        """
        Uniform sampling on simplices.

        Args:
            X (np.array): the base vectors
            simplices (np.array): the indexes of the simplex nodes
            X_vertices (np.array): array of neighbor vectors
            vertex_weights (np.array): weights of the neighbor vectors

        Returns:
            np.array: the random samples
        """
        if n_dim == 2:
            samples= self.deterministic_samples_dim_2(X,
                                                    simplices,
                                                    X_vertices,
                                                    vertex_weights)
            return samples

        if n_dim == 1:
            return self.deterministic_samples_dim_1(X, simplices)

        raise ValueError(f"deterministic calculations for dim {n_dim} are\
                            not implemented yet")

    def sample_simplex(self,
                X,
                *,
                indices,
                n_to_sample,
                base_weights=None,
                vertex_weights=None,
                X_vertices=None,
                simplex_weights=None):
        """
        Carry out the simplex sampling.

        Args:
            X (np.array): the base vectors
            n_to_sample (int): the number of samples to generate
            base_weights (np.array): the weights of the base vectors
            vertex_weights (np.array): weights of the neighbor vectors
            X_vertices (np.array): array of neighbor vectors
            simplex_weights (np.array): weights of the simplices

        Returns:
            np.array: the generated samples
        """

        if n_to_sample == 0:
            return np.zeros(shape=(0, X.shape[1]))

        n_dim = self.n_dim
        n_dim = np.min([n_dim, indices.shape[1]])

        if n_dim != self.n_dim:
            _logger.info("%s: The simplex order was updated from %d to %d",
                        self.__class__.__name__, self.n_dim, n_dim)

        _logger.info("%s: simplex sampling with n_dim %d",
                        self.__class__.__name__, n_dim)

        simplices = self.simplices(X,
                                n_to_sample,
                                base_weights=base_weights,
                                indices=indices,
                                X_vertices=X_vertices,
                                simplex_weights=simplex_weights,
                                n_dim=n_dim)

        if self.within_simplex_sampling == 'random':
            samples = random_samples_from_simplices(X,
                                                simplices,
                                                X_vertices,
                                                vertex_weights,
                                                self.random_state)
        elif self.within_simplex_sampling == 'deterministic':
            samples = self.deterministic_samples_from_simplices(X=X,
                                                                simplices=simplices,
                                                                X_vertices=X_vertices,
                                                                vertex_weights=vertex_weights,
                                                                n_dim=n_dim)
        else:
            msg = "Within simplex sampling strategy"\
                f" {self.within_simplex_sampling} is not implemented yet."
            raise ValueError(msg)
        return self.add_gaussian_noise(samples)



"""
This module implements the PDFOS method.
"""

import numpy as np
from numpy.linalg import LinAlgError

from scipy.special import expit

from sklearn.preprocessing import StandardScaler

from ..base import RandomStateMixin, coalesce
from ..base import MetricTensor
from ..base import OverSampling
from .._logger import logger
_logger= logger

__all__= ['PDFOS', 'FullRankTransformer', 'PDFOSKDE']

class FullRankTransformer:
    """
    Transforms data into a lower dimensional full rank covariance
    representation
    """
    def __init__(self, eps=0.002):
        """
        Constructor of the object.

        Args:
            eps (float): the tolerance
        """
        self.eps = eps
        self.mean = None
        self.transformation = None
        self.n_dim = None

    def fit(self, X):
        """
        Fits the transformer to the data

        Args:
            X (np.array): all vectors to fit on

        Returns:
            self: the fitted object
        """
        cov = np.cov(X, rowvar=False)
        eigw, eigv = np.linalg.eig(cov)

        eigw = np.real(eigw)
        eigv = np.real(eigv)

        eigv_f = eigv[:, eigw > self.eps]

        self.mean = np.mean(X, axis=0)
        self.transformation = eigv_f
        self.n_dim = eigv_f.shape[1]

        return self

    def transform(self, X):
        """
        Transforms the data

        Args:
            X (np.array): the data to transform

        Returns:
            np.array: the transformed data
        """
        if X is None:
            return None
        return np.dot(X - self.mean, self.transformation)

    def inverse_transform(self, X):
        """
        Inverse transformation

        Args:
            X (np.array): the data to inverse transform

        Returns:
            np.array: the inverse transformed data
        """
        return np.dot(X, self.transformation.T) + self.mean


class PDFOSKDE(RandomStateMixin):
    """
    The KDE in PDFOS
    """
    def __init__(self, metric_learning=None, random_state=None):
        """
        Constructor of the PDFOS KDE estimation

        Args:
            metric_learning (str/None): metric learning method
            random_state (int/None/np.random.RandomState): random state
                                                            initializer
        """
        RandomStateMixin.__init__(self, random_state)
        self.transformer = FullRankTransformer(eps=0.02)
        self.metric_learning = coalesce(metric_learning, 'cov_min')
        self.S = None # pylint: disable=invalid-name
        self.X_base = None

    def eq_5_0(self,
                sigma,
                m # pylint: disable=invalid-name
                ):
        """
        Eq (5) in the paper

        Args:
            sigma (float): the sigma value
            m (int): the dimensionality

        Returns:
            float: the value of the equation
        """
        return sigma**(-m)/((2.0*np.pi)**(m/2))

    def eq_8_vectorized(self,
                        X,
                        sigma,
                        S_inv # pylint: disable=invalid-name
                        ):
        """
        Eq (8) in the paper

        Args:
            X (np.array): all training vectors
            sigma (float): the sigma value
            S_inv (np.array): the inverse covariance matrix

        Returns:
            np.array: the result of the equation for all pairs of vectors
        """
        m = X.shape[1] # pylint: disable=invalid-name

        tmp= (X[:,None] - X)
        tmp = np.einsum('ijk,kl,lji -> ij', tmp, S_inv, tmp.T)

        numerator_9 = (np.sqrt(2)*sigma)**(-m)*expit(-(1/(4*sigma**2))*tmp)
        numerator_5 = sigma**(-m)*expit(-(1/(2*sigma**2))*tmp)
        denominator = ((2*np.pi)**(m/2))
        eq_9 = numerator_9 / denominator
        eq_5 = numerator_5 / denominator

        return eq_9 - 2 * eq_5

    def M(self, # pylint: disable=invalid-name
            sigma,
            X,
            S_inv # pylint: disable=invalid-name
            ):
        """
        Eq (7) in the paper

        Args:
            X (np.array): all training vectors
            sigma (float): the sigma value
            S_inv (np.array): the inverse covariance matrix

        Returns:
            float: the value of the equation
        """
        m = X.shape[1] # pylint: disable=invalid-name
        total = np.sum(self.eq_8_vectorized(X, sigma, S_inv))

        term_a = total/len(X)**2
        term_b = 2.0 * self.eq_5_0(sigma, m)/len(X)
        return term_a + term_b

    def find_best_sigma(self,
                        X,
                        n_optimize,
                        S_inv # pylint: disable=invalid-name
                        ):
        """
        Find the best sigma parameter.

        Args:
            X (np.array): all training vectors
            n_optimize (int): the number of vectors used for
                                the estimation
            S_inv (np.array): the inverse covariance matrix

        Returns:
            float: the best sigma value
        """
        # finding the best sigma parameter
        best_sigma = 0
        error = np.inf
        # the dataset is reduced to make the optimization more efficient
        domain = range(len(X))
        n_to_choose = min([len(X), n_optimize])
        X_reduced = X[self.random_state.choice(domain, # pylint: disable=invalid-name
                                               n_to_choose,
                                               replace=False)]

        # we suppose that the data is normalized, thus, this search space
        # should be meaningful
        for sigma in np.logspace(-5, 2, num=20):
            err = self.M(sigma, X_reduced, S_inv)
            if err < error:
                error = err
                best_sigma = sigma
        #_logger.info("%s: best sigma found %f",
        #                self.__class__.__name__, best_sigma)

        return best_sigma

    def fit(self,
                X,
                X_ml, # pylint: disable=invalid-name
                y_ml,
                n_optimize=100):
        """
        Fitting the kernel density estimator

        Args:
            X (np.array): the training vectors
            X_ml (np.array): potential training vectors for
                                    metric learning
            y_ml (np.array): potential target labels for metric
                                    learning
            n_optimize (int): the number of random samples used
                                for optimization

        Returns:
            obj: the fitted object
        """
        self.transformer.fit(X)
        X_trans = self.transformer.transform(X) # pylint: disable=invalid-name

        X_ml = self.transformer.transform(X_ml)

        metrict = MetricTensor(metric_learning_method=self.metric_learning)
        S_inv = metrict.tensor(X_ml, y_ml) # pylint: disable=invalid-name
        S = np.linalg.inv(S_inv) # pylint: disable=invalid-name

        best_sigma = self.find_best_sigma(X_trans, n_optimize, S_inv)

        self.S = best_sigma * S
        self.X_base = X_trans

        return self

    def sample(self, n_to_sample):
        """
        Draw a random sample from the fitted KDE

        Args:
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """
        samples_raw = self.random_state.multivariate_normal(np.repeat(0.0, self.X_base.shape[1]),
                                                            self.S,
                                                            size=n_to_sample)

        base_indices = self.random_state.choice(self.X_base.shape[0], n_to_sample)
        unique_indices, unique_counts = np.unique(base_indices, return_counts=True)
        samples = []
        loc = 0
        for idx, base_idx in enumerate(unique_indices):
            samples.append(self.X_base[base_idx] \
                            + samples_raw[loc:(loc + unique_counts[idx])])
            loc = loc + unique_counts[idx]
        samples = np.vstack(samples)
        return self.transformer.inverse_transform(samples)

class PDFOS(OverSampling):
    """
    References:
        * BibTex::

            @article{pdfos,
                    title = "PDFOS: PDF estimation based over-sampling for
                                imbalanced two-class problems",
                    journal = "Neurocomputing",
                    volume = "138",
                    pages = "248 - 259",
                    year = "2014",
                    issn = "0925-2312",
                    doi = "https://doi.org/10.1016/j.neucom.2014.02.006",
                    author = "Ming Gao and Xia Hong and Sheng Chen and Chris
                                J. Harris and Emad Khalaf",
                    keywords = "Imbalanced classification, Probability density
                                function based over-sampling, Radial basis
                                function classifier, Orthogonal forward
                                selection, Particle swarm optimisation"
                    }

    Notes:
        * Not prepared for low-rank data.
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_density_estimation]

    def __init__(self,
                 proportion=1.0,
                 *,
                 metric_learning=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            metric_learning (str/None): optional metric learning method 'ITML',
                                        'LSML'
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.metric_learning = metric_learning
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion,
                                           self.class_stats[self.maj_label],
                                           self.class_stats[self.min_label])

        if n_to_sample == 0:
            return self.return_copies(X, y, n_to_sample)

        # scaling the data to aid numerical stability
        scaler = StandardScaler()
        X_scaled = scaler.fit_transform(X) # pylint: disable=invalid-name

        X_min = X_scaled[y == self.min_label]

        # generating samples by kernel density estimation
        try:
            kde = PDFOSKDE(metric_learning=self.metric_learning,
                            random_state=self._random_state_init).fit(X_min,
                                                                        X_ml=X_scaled,
                                                                        y_ml=y)
        except LinAlgError as exc:
            return self.return_copies(X, y, f"kde fitting did not succeed {exc}")
        samples = kde.sample(n_to_sample)

        return (np.vstack([X, scaler.inverse_transform(samples)]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the SOI_CJ method.
"""
import itertools

import numpy as np

from ..base import coalesce, coalesce_dict
from ..base import (NearestNeighborsWithMetricTensor,
                            pairwise_distances_mahalanobis)
from ..base import OverSamplingSimplex

from .._logger import logger
_logger= logger

__all__= ['SOI_CJ']

class SOI_CJ(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @article{soi_cj,
                    author = {Sánchez, Atlántida I. and Morales, Eduardo and
                                Gonzalez, Jesus},
                    year = {2013},
                    month = {01},
                    pages = {},
                    title = {Synthetic Oversampling of Instances Using
                                Clustering},
                    volume = {22},
                    booktitle = {International Journal of Artificial
                                    Intelligence Tools}
                    }
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_uses_clustering,
                  OverSamplingSimplex.cat_sample_componentwise,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 method='jittering',
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of nearest neighbors in the SMOTE
                                sampling
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            method (str): 'interpolation'/'jittering'
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 1, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, 'proportion', 0)
        self.check_greater_or_equal(n_neighbors, 'n_neighbors', 1)
        self.check_isin(method, 'method', ['interpolation', 'jittering'])
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.method = method
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'method': ['interpolation', 'jittering']}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def rearrange_clusters(self, cluster_idx, cluster_jdx, intersection):
        """
        Move the difference into the bigger cluster.

        Args:
            cluster_idx (set): the first cluster
            cluster_jdx (set): the second cluster
            intersection (set): the intersection of the clusters

        Returns:
            set, set: the updated cluster_idx and cluster_jdx
        """
        # otherwise move the difference into the
        # bigger cluster
        if len(cluster_idx) > len(cluster_jdx):
            # removes the items that exist in both sets
            cluster_jdx.difference_update(intersection)
        else:
            cluster_idx.difference_update(intersection)

        return cluster_idx, cluster_jdx

    def update_clusters(self,
                            *,
                            X,
                            y,
                            cluster_idx,
                            cluster_jdx,
                            intersection,
                            nn_all,
                            nn_params):
        """
        Update two clusters during the merging.

        Args:
            X (np.array): the training vectors
            y (np.array): the target labels
            cluster_idx (set): the cluster at the index idx
            cluster_jdx (set): the cluster at the index jdx
            intersection (set): the intersection of the clusters
            nn_all (obj): fitted nearest neighbors object
            nn_params (dict): nearest neighbor parameters

        Returns:
            set, set: the updated clusters cluster_idx and cluster_jdx
        """
        # computing distance matrix
        distm = pairwise_distances_mahalanobis(X[list(cluster_idx)],
                                                Y=X[list(cluster_jdx)],
                                                tensor=nn_params.get('metric_tensor', None))
        # largest distance
        max_dist_pair = np.where(distm == np.max(distm))
        # elements with the largest distance
        max_i = X[list(cluster_idx)[max_dist_pair[0][0]]]
        max_j = X[list(cluster_jdx)[max_dist_pair[1][0]]]

        # finding midpoint and radius
        mid_point = (max_i + max_j)/2.0
        radius = np.linalg.norm(mid_point - max_i)

        # extracting points within the hypersphare of
        # radius "radius"
        ind = nn_all.radius_neighbors(mid_point.reshape(1, -1),
                                        radius,
                                        return_distance=False)

        if np.sum(y[ind[0].astype(int)] == self.min_label) > len(ind[0])/2:
            # if most of the covered elements come from the
            # minority class, merge clusters
            cluster_idx.update(cluster_jdx)
            cluster_jdx = set()
        else:
            cluster_idx, cluster_jdx = \
                self.rearrange_clusters(cluster_idx, cluster_jdx, intersection)

        return cluster_idx, cluster_jdx

    def init_clusters(self, y, indices):
        """
        Initialize the clusters.

        Args:
            y (np.array): all target labels
            indices (np.array): the neighborhood structure

        Returns:
            list(set): the initial clusters
        """
        clusters = []

        first_maj_index = np.argmax(y[indices] != self.min_label, axis=1)

        for idx in range(indices.shape[0]):
            # while the closest instance is from the minority class, adding it
            # to the cluster
            clusters.append(set(indices[idx, :first_maj_index[idx]].tolist()))

        return clusters

    def clustering(self, X, y, nn_params):
        """
        Implementation of the clustering technique described in the paper.

        Args:
            X (np.array): array of training instances
            y (np.array): target labels
            nn_params (dict): the nearest neighbor parameters

        Returns:
            list(set): list of minority clusters
        """
        nn_all= NearestNeighborsWithMetricTensor(n_jobs=self.n_jobs,
                                                 **nn_params)
        nn_all.fit(X)

        X_min = X[y == self.min_label]

        # extract nearest neighbors of all samples from the set of
        # minority samples
        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=len(X_min),
                                               n_jobs=self.n_jobs,
                                               **nn_params)
        nnmt.fit(X)
        indices = nnmt.kneighbors(X_min, return_distance=False)

        # initialize clusters by minority samples
        clusters = self.init_clusters(y, indices)

        # cluster merging phase
        is_intersection = True
        while is_intersection:
            is_intersection = False
            for idx, cluster_idx in enumerate(clusters):
                for jdx, cluster_jdx in enumerate(clusters[(idx + 1):]):
                    # computing intersection
                    intersection = cluster_idx.intersection(cluster_jdx)
                    if len(intersection) > 0:
                        is_intersection = True

                        cluster_idx, cluster_jdx = self.update_clusters(X=X, y=y,
                                                            cluster_idx=cluster_idx,
                                                            cluster_jdx=cluster_jdx,
                                                            intersection=intersection,
                                                            nn_all=nn_all,
                                                            nn_params=nn_params)

                        clusters[idx] = cluster_idx
                        clusters[jdx] = cluster_jdx


            clusters = [c for c in clusters if len(c) > 0]

        # returning non-empty clusters
        return clusters

    def sample_within_cluster_interpolation(self, X, cluster, n_to_sample):
        """
        Sample within a cluster by interpolation.

        Args:
            X (np.array): all training vectors
            cluster (set): the cluster indices
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """
        combinations = list(itertools.combinations(np.arange(len(cluster)), 2))
        indices = self.random_state.choice(np.arange(len(combinations)),
                                                        n_to_sample)

        combinations = np.array(combinations)[indices]
        coords = np.array(list(zip(*combinations)))

        X_0 = X[np.array(cluster)[coords[0]]] # pylint: disable=invalid-name
        X_1 = X[np.array(cluster)[coords[1]]] # pylint: disable=invalid-name
        samples = X_0 + (X_1 - X_0) \
                        * self.random_state.random_sample(size=X_0.shape)

        return samples

    def sampling_by_interpolation(self,
                                X,
                                clusters,
                                cluster_weights,
                                n_to_sample):
        """
        Generate samples by interpolation.

        Args:
            X (np.array): all training vectors
            clusters (list(set)): the cluster indices
            cluster_weights (np.array): the weights of the clusters
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """
        samples = np.zeros(shape=(0, X.shape[1]))
        cluster_indices = self.random_state.choice(np.arange(len(clusters)),
                                                    n_to_sample,
                                                    p=cluster_weights)
        cluster_unique, cluster_count = np.unique(cluster_indices,
                                                    return_counts=True)
        for idx, cluster_idx in enumerate(cluster_unique):
            samples_tmp = self.sample_within_cluster_interpolation(X,
                                                clusters[cluster_idx],
                                                cluster_count[idx])
            samples = np.vstack([samples, samples_tmp])

        return samples

    def cluster_stds(self, X, y, clusters):
        """
        Determine the cluster dependent stds.

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            clusters (list(set)): the cluster indices

        Returns:
            np.array: the stds for the clusters
        """
        std_min = np.std(X[y == self.min_label], axis=0)
        cluster_stds = []

        for cluster in clusters:
            std = np.std(X[cluster], axis=0)
            std = np.min(np.vstack([std_min, std]), axis=0)
            cluster_stds.append(std)

        return np.array(cluster_stds)

    def sample_within_cluster_jittering(self,
                                        *,
                                        X,
                                        cluster,
                                        n_to_sample,
                                        std,
                                        nn_params):
        """
        Sample within a cluster by interpolation.

        Args:
            X (np.array): all training vectors
            cluster (set): the cluster indices
            n_to_sample (int): the number of samples to generate
            std (np.array): the cluster specific standard deviations

        Returns:
            np.array: the generated samples
        """

        self.gaussian_component = {'sigmas': std}

        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=len(cluster),
                                               n_jobs=self.n_jobs,
                                               **nn_params)
        nnmt.fit(X[cluster])
        indices = nnmt.kneighbors(X[cluster], return_distance=False)

        samples = self.sample_simplex(X=X[cluster],
                                      indices=indices,
                                      n_to_sample=n_to_sample)

        #indices = self.random_state.choice(np.arange(len(cluster)),
        #                                        n_to_sample)
        #X_base = X[np.array(cluster)[indices]]
        #samples = X_base + self.random_state.normal(size=X_base.shape) * std

        return samples

    def sampling_by_jittering(self, *, X, y,
                            clusters,
                            cluster_weights,
                            n_to_sample,
                            nn_params):
        """
        Generate samples by jittering.

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            clusters (list(set)): the cluster indices
            cluster_weights (np.array): the weights of the clusters
            n_to_sample (int): the number of samples to generate
            nn_params (dict): the nearest neighbor parameters

        Returns:
            np.array: the generated samples
        """
        stds = self.cluster_stds(X, y, clusters)

        samples = np.zeros(shape=(0, X.shape[1]))

        cluster_indices = self.random_state.choice(np.arange(len(clusters)),
                                                    n_to_sample,
                                                    p=cluster_weights)
        cluster_unique, cluster_count = np.unique(cluster_indices,
                                                    return_counts=True)

        for idx, cluster_idx in enumerate(cluster_unique):
            X_samp = self.sample_within_cluster_jittering(X=X,
                                                cluster=clusters[cluster_idx],
                                                n_to_sample=cluster_count[idx],
                                                std=stds[cluster_idx],
                                                nn_params=nn_params)
            samples = np.vstack([samples, X_samp])

        return samples

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion,
                                           self.class_stats[self.maj_label],
                                           self.class_stats[self.min_label])

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
                    self.metric_tensor_from_nn_params(nn_params, X, y)

        # do the clustering
        _logger.info("%s: Executing clustering", self.__class__.__name__)
        clusters = self.clustering(X, y, nn_params)

        # filtering the clusters, at least two points in a cluster are needed
        # for both interpolation and jittering (due to the standard deviation)
        clusters_filtered = [list(c) for c in clusters if len(c) > 2]

        if len(clusters_filtered) == 0:
            return self.return_copies(X, y, "No clusters with at least 2 elements")

        # if there are clusters having at least 2 elements, do the sampling
        cluster_nums = [len(c) for c in clusters_filtered]
        cluster_weights = cluster_nums/np.sum(cluster_nums)

        _logger.info("%s: Executing sample generation", self.__class__.__name__)

        if self.method == 'interpolation':
            samples = self.sampling_by_interpolation(X,
                                                clusters_filtered,
                                                cluster_weights,
                                                n_to_sample)
        elif self.method == 'jittering':
            samples = self.sampling_by_jittering(X=X,
                                                 y=y,
                                                 clusters=clusters_filtered,
                                                 cluster_weights=cluster_weights,
                                                 n_to_sample=n_to_sample,
                                                 nn_params=nn_params)

        return (np.vstack([X, samples]),
                np.hstack([y, np.array([self.min_label]*len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'method': self.method,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
Testing the SMOTE_TomekLinks.
"""
import numpy as np

from smote_variants import SMOTE_TomekLinks

def test_specific():
    """
    Oversampler specific testing
    """
    obj = SMOTE_TomekLinks()

    X = np.array([[1, 2], [2, 2], [3, 4], [4, 4],
                    [5, 6], [6, 6], [8, 9]])
    y = np.array([0, 1, 0, 1, 0, 1, 0])

    assert obj.sample(X, y)[0].shape[0] > 0

    X_samp, _ = obj.return_results(np.zeros((0, 2)), np.array([]), X, y)

    assert X_samp.shape[0] == X.shape[0]



"""
This module implements the ADASYN technique.
"""

import numpy as np

from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from ..base import coalesce, coalesce_dict

from .._logger import logger
_logger= logger

__all__= ['ADASYN']

class ADASYN(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @inproceedings{adasyn,
                          author={He, H. and Bai, Y. and Garcia,
                                    E. A. and Li, S.},
                          title={{ADASYN}: adaptive synthetic sampling
                                    approach for imbalanced learning},
                          booktitle={Proceedings of IJCNN},
                          year={2008},
                          pages={1322--1328}
                        }
    """

    categories = [OverSamplingSimplex.cat_sample_ordinary,
                  OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_borderline,
                  OverSamplingSimplex.cat_density_based,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 n_neighbors=5,
                 beta=1.0,
                 *,
                 d_th=0.9,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 proportion=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            n_neighbors (int): control parameter of the nearest neighbor
                                component
            d_th (float): tolerated deviation level from balancedness
            beta (float): target level of balancedness, same as proportion
                            in other techniques
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling params
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
            proportion (float): same as beta, for convenience
        """
        nn_params = coalesce(nn_params, {})

        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state, checks=None)

        self.check_greater_or_equal(n_neighbors, 'n_neighbors', 1)
        self.check_greater_or_equal(d_th, 'd_th', 0)
        self.check_greater_or_equal(beta, 'beta', 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.n_neighbors = n_neighbors
        self.d_th = d_th
        self.beta = proportion or beta
        self.proportion = self.beta
        self.nn_params = nn_params
        self.n_jobs = n_jobs

    @classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'n_neighbors': [3, 5, 7, 9],
                                  'd_th': [0.9],
                                  'proportion': [2.0, 1.5, 1.0, 0.75, 0.5, 0.25]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.beta)

        if n_to_sample == 0:
            return self.return_copies(X, y, "no need for sampling")

        # extracting minority samples
        X_min = X[y == self.min_label]

        # checking if sampling is needed
        m_min = len(X_min)
        m_maj = len(X) - m_min

        d = float(m_min)/m_maj # pylint: disable=invalid-name
        if d > self.d_th: # pylint: disable=invalid-name
            _logger.warning("%s: d > d_th",
                            self.__class__.__name__)
            return X.copy(), y.copy()

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        # fitting nearest neighbors model to all samples
        n_neigh = min([len(X_min), self.n_neighbors+1])
        nearestn= NearestNeighborsWithMetricTensor(n_neighbors=n_neigh,
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nearestn.fit(X)
        indices = nearestn.kneighbors(X_min, return_distance=False)

        # determining the distribution of points to be generated
        r = [] #pylint: disable=invalid-name
        for _, row in enumerate(indices):
            r.append(sum(y[row[1:]] ==
                         self.maj_label)/self.n_neighbors)

        r = np.array(r) #pylint: disable=invalid-name

        if np.sum(r) == 0:
            _logger.warning("%s: not enough samples close to majority ones "\
                            "for oversampling",
                            self.__class__.__name__)
            return X.copy(), y.copy()

        r = r/np.sum(r) #pylint: disable=invalid-name

        # fitting nearest neighbors models to minority samples
        n_neigh = min([len(X_min), self.n_neighbors + 1])
        nearestn= NearestNeighborsWithMetricTensor(n_neighbors=n_neigh,
                                                        n_jobs=self.n_jobs,
                                                        **nn_params)
        nearestn.fit(X_min)
        indices = nearestn.kneighbors(X_min, return_distance=False)

        samples = self.sample_simplex(X=X_min,
                                        indices=indices,
                                        n_to_sample=n_to_sample,
                                        base_weights=r)

        return (np.vstack([X, samples]),
                np.hstack([y, np.hstack([self.min_label]*int(n_to_sample))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'n_neighbors': self.n_neighbors,
                'd_th': self.d_th,
                'beta': self.beta,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                'proportion': self.proportion,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the ROSE method.
"""

import numpy as np

from ..base import OverSampling

from .._logger import logger
_logger= logger

__all__= ['ROSE']

class ROSE(OverSampling):
    """
    References:
        * BibTex::

            @Article{rose,
                    author="Menardi, Giovanna
                    and Torelli, Nicola",
                    title="Training and assessing classification rules with
                            imbalanced data",
                    journal="Data Mining and Knowledge Discovery",
                    year="2014",
                    month="Jan",
                    day="01",
                    volume="28",
                    number="1",
                    pages="92--122",
                    issn="1573-756X",
                    doi="10.1007/s10618-012-0295-5",
                    url="https://doi.org/10.1007/s10618-012-0295-5"
                    }

    Notes:
        * It is not entirely clear if the authors propose kernel density
            estimation or the fitting of simple multivariate Gaussians
            on the minority samples. The latter seems to be more likely,
            I implement that approach.
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_sample_componentwise]

    def __init__(self,
                 proportion=1.0,
                 *,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, 'proportion', 0.0)

        self.proportion = proportion

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion,
                                           self.class_stats[self.maj_label],
                                           self.class_stats[self.min_label])

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        # Estimating the H matrix
        std = np.std(X_min, axis=0)
        n, d = X.shape # pylint: disable=invalid-name
        H = std*(4.0/((d + 1)*n))**(1.0/(d + 4)) # pylint: disable=invalid-name

        base_indices = self.random_state.choice(np.arange(X_min.shape[0]),
                                                n_to_sample)
        base_vectors = X_min[base_indices]
        random = self.random_state.normal(size=base_vectors.shape)

        samples = base_vectors + random * H

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                **OverSampling.get_params(self)}



"""
This module implements the SVM_balance method.
"""

import numpy as np

from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC
from sklearn.model_selection import cross_val_score

from ..base import coalesce, coalesce_dict
from ..base import OverSampling
from ._smote import SMOTE

from .._logger import logger
_logger= logger

__all__= ['SVM_balance']

class SVM_balance(OverSampling):
    """
    References:
        * BibTex::

            @article{svm_balance,
                     author = {Farquad, M.A.H. and Bose, Indranil},
                     title = {Preprocessing Unbalanced Data Using Support
                                Vector Machine},
                     journal = {Decis. Support Syst.},
                     issue_date = {April, 2012},
                     volume = {53},
                     number = {1},
                     month = apr,
                     year = {2012},
                     issn = {0167-9236},
                     pages = {226--233},
                     numpages = {8},
                     url = {http://dx.doi.org/10.1016/j.dss.2012.01.016},
                     doi = {10.1016/j.dss.2012.01.016},
                     acmid = {2181554},
                     publisher = {Elsevier Science Publishers B. V.},
                     address = {Amsterdam, The Netherlands, The Netherlands},
                     keywords = {COIL data, Hybrid method, Preprocessor, SVM,
                                    Unbalanced data},
                    }
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_uses_classifier,
                  OverSampling.cat_changes_majority,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors in the SMOTE sampling
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}

        super().__init__(random_state=random_state)

        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.ss_params = coalesce_dict(ss_params, ss_params_default)
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        X, y = SMOTE(proportion=self.proportion,
                     n_neighbors=self.n_neighbors,
                     nn_params=self.nn_params,
                     ss_params=self.ss_params,
                     n_jobs=self.n_jobs,
                     random_state=self._random_state_init).sample(X, y)

        validator = np.min([5, sum(y == self.min_label)])

        scaler = StandardScaler()
        X_norm = scaler.fit_transform(X) # pylint: disable=invalid-name

        C_params = [0.01, 0.1, 1.0, 10.0] # pylint: disable=invalid-name
        best_score = 0
        best_C = 0.01 # pylint: disable=invalid-name

        for C in C_params: # pylint: disable=invalid-name
            _logger.info("%s: Evaluating SVM with C=%f",
                                    self.__class__.__name__, C)

            svc = SVC(C=C, kernel='rbf', gamma='auto')
            score = np.mean(cross_val_score(svc, X_norm, y, cv=validator))
            if score > best_score:
                best_score = score
                best_C = C # pylint: disable=invalid-name

        svc = SVC(C=best_C, kernel='rbf', gamma='auto')

        svc.fit(X_norm, y)

        return X, svc.predict(X_norm)

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'ss_params': self.ss_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
Testing the visualization tools.
"""
import numpy as np

from smote_variants.visualization import (vector_labeling,
                                            plot_raw,
                                            plot_oversampled,
                                            plot_sampling_raw,
                                            plot_comparison_raw,
                                            relabel_y,
                                            set_seaborn_theme)

from smote_variants import SMOTE
from smote_variants.datasets import load_illustration_2_class, load_illustration_3_class

dataset = load_illustration_2_class()

X = dataset['data']
y = dataset['target']

X_samp, y_samp = SMOTE().sample(X, y)


def test_style():
    """
    Testing the setting of the color palette
    """

    assert len(set_seaborn_theme(2)) != len(set_seaborn_theme(5))


def test_relabeling():
    """
    Test the relabeling
    """
    dataset_3 = load_illustration_3_class()

    assert len(np.unique(relabel_y(dataset_3['target']))) == 3

def test_vector_labeling():
    """
    Testing the vector labeling
    """

    assert len(vector_labeling(X, y, X_samp, y_samp)) > 0

def test_plot_raw():
    """
    Testing the raw plotting
    """

    assert plot_raw(X, y) is not None

def test_plot_oversampled():
    """
    Testing the oversampled dataset plotting
    """

    assert plot_oversampled(X, y, X_samp, y_samp) is not None

def test_plot_sampling_raw():
    """
    Testing the plot sampling raw function
    """

    assert plot_sampling_raw(X=X,
                             y=y,
                             X_samp=X_samp,
                             y_samp=y_samp,
                             title="SMOTE") is not None

def test_plot_comparison_raw():
    """
    Testing the plot_comparison_raw function
    """
    X_samp1, y_samp1 = SMOTE(n_neighbors=15).sample(X, y) # pylint: disable=invalid-name

    fig = plot_comparison_raw(X=X, y=y, X_samp0=X_samp, y_samp0=y_samp,
                                X_samp1=X_samp1, y_samp1=y_samp1,
                                title0='SMOTE', title1='SMOTE15')
    assert fig is not None



"""
This module implements the SMOTE_IPF method.
"""

import numpy as np

from sklearn.model_selection import StratifiedKFold

from ..base import coalesce, coalesce_dict
from ..base import OverSampling
from ._smote import SMOTE
from ..base import instantiate_obj

from .._logger import logger
_logger= logger

__all__= ['SMOTE_IPF']

class SMOTE_IPF(OverSampling):
    """
    References:
        * BibTex::

            @article{smote_ipf,
                        title = "SMOTE–IPF: Addressing the noisy and borderline
                                    examples problem in imbalanced
                                    classification by a re-sampling method
                                    with filtering",
                        journal = "Information Sciences",
                        volume = "291",
                        pages = "184 - 203",
                        year = "2015",
                        issn = "0020-0255",
                        doi = "https://doi.org/10.1016/j.ins.2014.08.051",
                        author = "José A. Sáez and Julián Luengo and Jerzy
                                    Stefanowski and Francisco Herrera",
                        keywords = "Imbalanced classification,
                                        Borderline examples,
                                        Noisy data,
                                        Noise filters,
                                        SMOTE"
                        }
    """

    categories = [OverSampling.cat_changes_majority,
                  OverSampling.cat_uses_classifier,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_folds=9,
                 k=3,
                 p=0.01,
                 voting='majority',
                 classifier=('sklearn.tree',
                                'DecisionTreeClassifier',
                                {'random_state': 2}),
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            n_neighbors (int): number of neighbors in SMOTE sampling
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_folds (int): the number of partitions
            k (int): used in stopping condition
            p (float): percentage value ([0,1]) used in stopping condition
            voting (str): 'majority'/'consensus'
            classifier (obj): classifier object
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}

        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater_or_equal(n_folds, "n_folds", 2)
        self.check_greater_or_equal(k, "k", 1)
        self.check_greater_or_equal(p, "p", 0)
        self.check_isin(voting, "voting", ['majority', 'consensus'])
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.ss_params = coalesce_dict(ss_params, ss_params_default)
        self.params = {'n_folds': n_folds,
                        'k': k,
                        'p': p,
                        'voting': voting,
                        'classifier': classifier}
        self.classifier_obj = instantiate_obj(classifier)
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        classifiers = [('sklearn.tree', 'DecisionTreeClassifier', {'random_state': 2})]
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'n_folds': [9],
                                  'k': [3],
                                  'p': [0.01],
                                  'voting': ['majority', 'consensus'],
                                  'classifier': classifiers}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        # do SMOTE sampling
        X_samp, y_samp = SMOTE(proportion=self.proportion,
                               n_neighbors=self.n_neighbors,
                               nn_params=self.nn_params,
                               ss_params=self.ss_params,
                               n_jobs=self.n_jobs,
                               random_state=self._random_state_init).sample(X, y)

        n_folds = min([self.params['n_folds'], np.sum(y == self.min_label)])

        condition = 0
        while condition < self.params['k']:
            # validating the sampled dataset
            validator = StratifiedKFold(n_folds)
            predictions = []
            for train_index, _ in validator.split(X_samp, y_samp):
                self.classifier_obj.fit(X_samp[train_index], y_samp[train_index])
                predictions.append(self.classifier_obj.predict(X_samp))

            # do decision based on one of the voting schemes
            if self.params['voting'] == 'majority':
                pred_votes = (np.mean(predictions, axis=0) > 0.5).astype(int)
                to_remove = np.where(np.not_equal(pred_votes, y_samp))[0]
            elif self.params['voting'] == 'consensus':
                pred_votes = (np.mean(predictions, axis=0) > 0.5).astype(int)
                sum_votes = np.sum(predictions, axis=0)
                to_remove = np.where(np.logical_and(np.not_equal(
                    pred_votes, y_samp), np.equal(sum_votes, self.params['n_folds'])))[0]

            # delete samples incorrectly classified
            _logger.info("%s: Removing %d elements",
                    self.__class__.__name__ , len(to_remove))
            X_samp = np.delete(X_samp, to_remove, axis=0)
            y_samp = np.delete(y_samp, to_remove)

            # if the number of samples removed becomes small or k iterations
            # were done quit
            check = int(len(to_remove) < len(X_samp) * self.params['p'])
            condition = (condition + 1) * check

        return X_samp, y_samp

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'ss_params': self.ss_params,
                'n_folds': self.params['n_folds'],
                'k': self.params['k'],
                'p': self.params['p'],
                'voting': self.params['voting'],
                'n_jobs': self.n_jobs,
                'classifier': self.params['classifier'],
                **OverSampling.get_params(self)}



"""
This module implements the G_SMOTE method.
"""

import numpy as np

from sklearn.metrics import pairwise_distances

from ..base import coalesce, coalesce_dict
from ..base._simplexsampling import array_array_index
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['G_SMOTE']

class G_SMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{g_smote,
                            author={Sandhan, T. and Choi, J. Y.},
                            booktitle={2014 22nd International Conference on
                                        Pattern Recognition},
                            title={Handling Imbalanced Datasets by Partially
                                    Guided Hybrid Sampling for Pattern
                                    Recognition},
                            year={2014},
                            volume={},
                            number={},
                            pages={1449-1453},
                            keywords={Gaussian processes;learning (artificial
                                        intelligence);pattern classification;
                                        regression analysis;sampling methods;
                                        support vector machines;imbalanced
                                        datasets;partially guided hybrid
                                        sampling;pattern recognition;real-world
                                        domains;skewed datasets;dataset
                                        rebalancing;learning algorithm;
                                        extremely low minority class samples;
                                        classification tasks;extracted hidden
                                        patterns;support vector machine;
                                        logistic regression;nearest neighbor;
                                        Gaussian process classifier;Support
                                        vector machines;Proteins;Pattern
                                        recognition;Kernel;Databases;Gaussian
                                        processes;Vectors;Imbalanced dataset;
                                        protein classification;ensemble
                                        classifier;bootstrapping;Sat-image
                                        classification;medical diagnoses},
                            doi={10.1109/ICPR.2014.258},
                            ISSN={1051-4651},
                            month={Aug}}

    Notes:
        * the non-linear approach is inefficient
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_sample_componentwise,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 method='linear',
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors in nearest neighbors
                                component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): the simplex sampling parameters
            method (str): 'linear'/'non-linear_2.0' - the float can be any
                                number: standard deviation in the
                                Gaussian-kernel
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        nn_params = coalesce(nn_params, {})

        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state, checks={'min_n_dim': 2})
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)

        self.sigma = None
        if method != 'linear' and not method.startswith('non-linear'):
            raise ValueError(f"{self.__class__.__name__}: method parameters "\
                                f"{method} is not supported")
        if method.startswith('non-linear'):
            self.sigma = float(method.split('_')[-1])
            if self.sigma <= 0:
                raise ValueError(f"{self.__class__.__name__}: Non-positive "\
                                f"non-linear parameter {self.sigma} is not "\
                                "supported")

        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = nn_params
        self.method = method
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'method': ['linear', 'non-linear_0.1',
                                             'non-linear_1.0',
                                             'non-linear_2.0']}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def determine_H(self, X_min): # pylint: disable=invalid-name
        """
        Determine the pricinpal direction H and the kernel

        Args:
            X_min (np.array): the minority vectors

        Returns:
            np.array, callable: the principal direction and the kernel
        """
        if self.method == 'linear':
            # finding H_l by linear decomposition
            cov = np.cov(X_min, rowvar=False)
            eigw, eigv = np.linalg.eig(cov)
            H = eigv[np.argmax(eigw)] # pylint: disable=invalid-name
        else:
            # building a non-linear kernel matrix and finding H_n by its
            # decomposition

            # I believe there are typos in the paper where the functional
            # form of the Gaussian-kernel is specified, the negative sign
            # and the square of the norm has been added.

            kernel_matrix = pairwise_distances(X_min)**2
            kernel_matrix = kernel_matrix/(2.0*self.sigma**2)
            kernel_matrix = np.exp(kernel_matrix)
            eigw, eigv = np.linalg.eig(kernel_matrix)
            H = eigv[np.argmax(eigw)] # pylint: disable=invalid-name

        return H

    def update_indices_by_angles(self,
                                    X,
                                    indices,
                                    H, # pylint: disable=invalid-name
                                    X_min):
        """
        Rearrange the neighborhood indices according to the angles

        Args:
            X (np.array): the base vectors
            indices (np.array): the neighborhoods
            H (np.array): the principal direction
            X_min (np.array): the minority samples used to create the
                                feature space representation in the non-linear
                                case
        Returns:
            np.array: the ordered neighborhoods
        """
        if self.method == 'linear':
            thetas = self.angles(X, indices, H)
        else:
            thetas = self.angles_non_linear(X, indices, H, X_min)
        thetas_argsort = thetas.argsort()
        indices_ordered = array_array_index(indices[:,1:], thetas_argsort)
        thetas_ordered = array_array_index(thetas, thetas_argsort)
        return indices_ordered, thetas_ordered

    def angles(self,
                X,
                indices,
                H # pylint: disable=invalid-name
                ):
        """
        Calculate the angles for each vector in each neighborhood
        in the linear case.

        Args:
            X (np.array): the base vectors
            indices (np.array): the neighborhoods
            H (np.array): the principal direction

        Returns:
            np.array: the angles for each vector in each neighborhood
        """
        P = X[indices[:, 1:]] - X[indices[:, 0]][:, None] # pylint: disable=invalid-name
        inner_product = np.abs(np.einsum('ijk,k->ij', P, H))
        norms = np.linalg.norm(P, axis=2) * np.linalg.norm(H)
        norms[norms == 0.0] = 1e-5

        return np.arccos(inner_product / norms)

    def angles_non_linear(self,
                            X,
                            indices,
                            H, # pylint: disable=invalid-name
                            X_min):
        """
        Calculate the angles in kernel space for each vector in each
        neighborhood in a vectorized form.

        Args:
            X (np.array): base vectors
            indices (np.array): the indices specifying the neighborhoods
            H (np.array): the principal direction
            X_min (np.array): the minority samples used for creating the
                                feature space representation

        Returns:
            np.array: the angles for each vector in each neighborhood
        """
        diff = X[indices[:, 1:]] - X[indices[:, 0]][:, None]
        P = diff[:, None] - X_min[:, None] # pylint: disable=invalid-name
        gram = -np.linalg.norm(P, axis=3)**2 / (2.0 * self.sigma**2)
        inner_product = np.abs(np.einsum('ikj,k->ij', gram, H))
        norm_H = np.linalg.norm(H) # pylint: disable=invalid-name
        norm_gram = np.linalg.norm(gram, axis=1)

        return np.arccos(inner_product / (norm_H * norm_gram))

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        # fitting nearest neighbors model
        n_neighbors = min([len(X_min), self.n_neighbors+1])

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nnmt.fit(X_min)
        ind = nnmt.kneighbors(X_min, return_distance=False)

        H = self.determine_H(X_min) # pylint: disable=invalid-name

        # generating samples
        indices_ordered, weights = \
                self.update_indices_by_angles(X_min, ind, H, X_min)

        indices_ordered = np.vstack([np.arange(len(indices_ordered)),
                                    indices_ordered[:,:(self.n_dim-1)].T]).T

        weights = np.vstack([np.repeat(1.0, len(weights)),
                            weights[:,:(self.n_dim-1)].T]).T

        samples = self.sample_simplex(X=X_min,
                                        indices=indices_ordered,
                                        n_to_sample=n_to_sample)
        # the weighted case
        #samples = self.sample_simplex(X=X_min,
        #                                indices=indices_ordered,
        #                                n_to_sample=n_to_sample,
        #                                X_vertices=X_min,
        #                                simplex_weights=weights)

        #while len(samples) < n_to_sample:
        #    idx = self.random_state.randint(len(X_min))
            # calculating difference vectors from all neighbors

            #P = X_min[ind[idx][1:]] - X_min[idx]
            #if self.method == 'linear':
            #    # calculating angles with the principal direction
            #    thetas = np.array([self.angle(P, n, H) for n in range(len(P))])
            #else:
            #    thetas = []
            #    # calculating angles of the difference vectors and the
            #    # principal direction in feature space
            #    for n in range(len(P)):
            #        # calculating representation in feature space
            #        feature_vector = np.array([kernel(X_min[k], P[n]) for k in range(len(X_min))])
            #        dp = np.dot(H, feature_vector)
            #        denom = np.linalg.norm(feature_vector)*np.linalg.norm(H)
            #        thetas.append(np.arccos(np.abs(dp)/denom))
            #    thetas = np.array(thetas)

            # using the neighbor with the difference along the most similar
            # direction to the principal direction of the data
            #n = np.argmin(thetas)
        #    X_a = X_min[idx]
            #X_b = X_min[ind[idx][1:][n]]
        #    X_b = X_min[indices_ordered[idx][0]]
        #    samples.append(self.sample_between_points_componentwise(X_a, X_b))

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'method': self.method,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This moduel imports the multiclass oversampling capabilities.
"""
import inspect

import numpy as np

from ..base import StatisticsMixin, instantiate_obj
from .._logger import logger

_logger = logger

# exported names
__all__ = ['MulticlassOversampling']

def proportion_1_vs_many(stats,
                        labels,
                        maj_label,
                        class_idx):
    """
    Determines the proportion to sample for the 1_vs_many strategy.

    Args:
        stats (dict): class statistics
        labels (list): class labels
        maj_label (int): the majority label
        class_idx (int): the class index to determine the proportion for

    Returns:
        float: the proportion
    """
    to_gen = stats[maj_label] - stats[labels[class_idx]]
    to_gen_to_all = stats[maj_label] - stats[labels[class_idx]]

    return to_gen/to_gen_to_all

def proportion_1_vs_many_succ(stats,
                                labels,
                                maj_label,
                                class_idx):
    """
    Determines the proportion to sample for the 1_vs_many_successive strategy.

    Args:
        stats (dict): class statistics
        labels (list): class labels
        maj_label (int): the majority label
        class_idx (int): the class index to determine the proportion for

    Returns:
        float: the proportion
    """
    n_majority = stats[maj_label]
    n_class_i = stats[labels[class_idx]]
    num_to_generate = n_majority - n_class_i

    num_to_gen_to_all = class_idx * n_majority - n_class_i

    return num_to_generate/num_to_gen_to_all

def has_proportion_parameter(oversampler):
    """
    Checks if an oversampler object has proportion parameters

    Args:
        oversampler: oversampler object

    Returns:
        bool: True if the object has 'proportion' parameter
    """
    return 'proportion' in \
            list(inspect.signature(oversampler.__class__).parameters.keys())

class MulticlassOversampling(StatisticsMixin):
    """
    Carries out multiclass oversampling

    Example::

        import smote_variants as sv
        import sklearn.datasets as datasets

        dataset= datasets.load_wine()

        oversampler= sv.MulticlassOversampling(sv.distance_SMOTE())

        X_samp, y_samp= oversampler.sample(dataset['data'], dataset['target'])
    """

    def __init__(self,
                 oversampler='SMOTE',
                 oversampler_params=None,
                 strategy="eq_1_vs_many_successive"):
        """
        Constructor of the multiclass oversampling object

        Args:
            oversampler (obj): an oversampling object
            strategy (str/obj): a multiclass oversampling strategy, currently
                                'eq_1_vs_many_successive' or
                                'equalize_1_vs_many'
        """
        self.oversampler = oversampler
        if oversampler_params is None:
            oversampler_params = {}
        self.oversampler_params = oversampler_params
        self.strategy = strategy

        if not has_proportion_parameter(instantiate_obj(('smote_variants',
                                                    self.oversampler,
                                                    self.oversampler_params))):
            raise ValueError((f"Multiclass oversampling strategy {self.strategy}"
                       " cannot be used with oversampling techniques without"
                       " proportion parameter"))

        if self.strategy not in ['eq_1_vs_many_successive',
                                        'equalize_1_vs_many']:
            message = "Multiclass oversampling startegy %s not implemented."
            message = message % self.strategy
            raise ValueError(message)

    def sample_equalize_1_vs_many(self, X, y):
        """
        Does the sample generation by oversampling each minority class to the
        cardinality of the majority class using all original samples in each
        run.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        _logger.info("%s: %s", self.__class__.__name__,
                f"Running multiclass oversampling with strategy {self.strategy}")

        # extract class label statistics
        self.class_label_statistics(y)

        # sort labels by number of samples
        labels = self.class_stats.keys()
        labels = sorted(labels, key=lambda x: -self.class_stats[x])

        # dict to store the results
        results = {}
        results[labels[0]] = X[y == labels[0]]

        # running oversampling for all minority classes against all oversampled
        # classes
        for class_idx in range(1, len(labels)):
            _logger.info("%s: %s", self.__class__.__name__,
                    f"Sampling minority class with label: {labels[class_idx]}")

            # prepare data to pass to oversampling
            X_train = np.vstack([X[y != labels[class_idx]],
                                    X[y == labels[class_idx]]])
            y_train = np.hstack([np.repeat(0, np.sum(y != labels[class_idx])),
                                 np.repeat(1, np.sum(y == labels[class_idx]))])

            # prepare parameters by properly setting the proportion value
            params = self.oversampler_params.copy()
            params['proportion'] = proportion_1_vs_many(self.class_stats,
                                                        labels,
                                                        labels[0],
                                                        class_idx)

            # instantiating new oversampling object with the proper proportion
            # parameter
            oversampler = instantiate_obj(('smote_variants',
                                            self.oversampler,
                                            params))

            # executing the sampling
            X_samp, y_samp = oversampler.sample(X_train, y_train)

            # registering the newly oversampled minority class in the output
            # set
            results[labels[class_idx]] = X_samp[len(X_train):][y_samp[len(X_train):] == 1]

        # constructing the output set
        X_final = results[labels[1]]
        y_final = np.repeat(labels[1], len(results[labels[1]]))

        for class_idx in range(2, len(labels)):
            X_final = np.vstack([X_final, results[labels[class_idx]]])
            y_final = np.hstack([y_final, np.repeat(labels[class_idx],
                                            len(results[labels[class_idx]]))])

        return np.vstack([X, X_final]), np.hstack([y, y_final])

    def sample_equalize_1_vs_many_successive(self, X, y):
        """
        Does the sample generation by oversampling each minority class
        successively to the cardinality of the majority class,
        incorporating the results of previous oversamplings.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        _logger.info("%s: %s", self.__class__.__name__,
            f"Running multiclass oversampling with strategy {self.strategy}")

        # extract class label statistics
        self.class_label_statistics(y)

        # sort labels by number of samples
        labels = self.class_stats.keys()
        labels = sorted(labels, key=lambda x: -self.class_stats[x])

        # determining the majority class data
        X_maj = X[y == labels[0]]

        # dict to store the results
        results = {labels[0]: X_maj}

        # running oversampling for all minority classes against all
        # oversampled classes
        for class_idx in range(1, len(labels)):
            _logger.info("%s: %s", self.__class__.__name__,
                f"Sampling minority class with label: {labels[class_idx]}")

            # prepare data to pass to oversampling
            X_train = np.vstack([X_maj, X[y == labels[class_idx]]])
            y_train = np.hstack([np.repeat(0, len(X_maj)),
                                 np.repeat(1, np.sum(y == labels[class_idx]))])

            # prepare parameters by properly setting the proportion value
            params = self.oversampler_params.copy()
            params['proportion'] = proportion_1_vs_many_succ(self.class_stats,
                                                                labels,
                                                                labels[0],
                                                                class_idx)

            # executing the sampling
            X_samp, y_samp = instantiate_obj(('smote_variants',
                                                self.oversampler,
                                                params)).sample(X_train, y_train)

            # adding the newly oversampled minority class to the majority data
            X_maj = np.vstack([X_maj, X_samp[y_samp == 1]])

            # registaring the newly oversampled minority class in the output
            # set
            result_mask = y_samp[len(X_train):] == 1
            results[labels[class_idx]] = X_samp[len(X_train):][result_mask]

        # constructing the output set
        X_final = results[labels[1]]
        y_final = np.repeat(labels[1], len(results[labels[1]]))

        for class_idx in range(2, len(labels)):
            X_final = np.vstack([X_final, results[labels[class_idx]]])
            y_final = np.hstack([y_final, np.repeat(labels[class_idx],
                                            len(results[labels[class_idx]]))])

        return np.vstack([X, X_final]), np.hstack([y, y_final])

    def sample(self, X, y):
        """
        Does the sample generation according to the oversampling strategy.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        if self.strategy == "eq_1_vs_many_successive":
            return self.sample_equalize_1_vs_many_successive(X, y)

        return self.sample_equalize_1_vs_many(X, y)

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the multiclass oversampling object
        """
        _ = deep
        return {'oversampler': self.oversampler,
                'oversampler_params': self.oversampler_params,
                'strategy': self.strategy}
